from pydantic import (
    BaseModel,
    ConfigDict,
    field_serializer,
)
from typing import Any


class PublicAnswerSerializer(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    id: int
    text: str


class PublicQuestionSerializer(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    id: int
    text: str
    answers: Any

    @field_serializer("answers")
    def serialize_answers(self, answers: Any) -> list[dict]:
        return [
            PublicAnswerSerializer.model_validate(answer).model_dump()
            for answer in answers.all()
        ]


class PublicQuizSerializer(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    id: int
    title: str
    questions: Any

    @field_serializer("questions")
    def serialize_questions(self, questions: Any) -> list[dict]:
        return [
            PublicQuestionSerializer.model_validate(question).model_dump()
            for question in questions.all()
        ]



import uuid

from django.db import models


class SessionState(models.TextChoices):
    PENDING = "PENDING", "Pending"
    PROCESSING = "PROCESSING", "Processing"
    COMPLETED = "COMPLETED", "Completed"
    FAILED = "FAILED", "Failed"


class Session(models.Model):
    session_id = models.CharField(max_length=255, unique=True)
    created_at = models.DateTimeField(auto_now_add=True)
    jwt_token = models.TextField()
    state = models.CharField(
        max_length=255,
        choices=SessionState.choices,
        default=SessionState.PENDING,
    )

    def save(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        if not self.session_id:
            self.session_id = str(uuid.uuid4())
        super().save(*args, **kwargs)



from django_email_learning.ports.delivery_queue_protocol import DeliveryQueueProtocol
from django_email_learning.models import DeliverySchedule, DeliveryStatus
from django.db import transaction
from django.utils import timezone
from typing import Iterator


class DatabaseDeliveryQueue(DeliveryQueueProtocol):
    ITERATOR_BATCH_SIZE = 50

    def __init__(self) -> None:
        self._task_iterator = self.get_next_batch(limit=self.ITERATOR_BATCH_SIZE)

    def get_next_batch(self, limit: int) -> Iterator[DeliverySchedule]:
        with transaction.atomic():
            # Get IDs of ready tasks while locked
            task_ids = list(
                DeliverySchedule.objects.select_for_update(skip_locked=True)  # type: ignore[misc]
                .filter(status=DeliveryStatus.SCHEDULED, time__lte=timezone.now())[
                    :limit
                ]
                .values_list("id", flat=True)
            )

            if not task_ids:
                return iter([])

            # Update status
            DeliverySchedule.objects.filter(id__in=task_ids).update(
                status=DeliveryStatus.PROCESSING
            )

        # Return fresh objects outside transaction
        return (
            DeliverySchedule.objects.filter(id__in=task_ids)
            .select_related("delivery__enrollment__learner", "delivery__course_content")
            .iterator()
        )

    def next_task(self) -> DeliverySchedule | None:
        try:
            return next(self._task_iterator)
        except StopIteration:
            self._task_iterator = self.get_next_batch(limit=self.ITERATOR_BATCH_SIZE)
            try:
                return next(self._task_iterator)
            except StopIteration:
                return None



from django.urls import path
from .views import EditTextView

app_name = "django_email_learning"

urlpatterns = [
    path(
        "organizations/<int:organization_id>/edit-text/",
        EditTextView.as_view(),
        name="edit_text",
    ),
]



from django.http import JsonResponse
from django.db.models import QuerySet
from typing import Any


class PaginatedApiMixin:
    def get(self, request, *args, **kwargs) -> JsonResponse:  # type: ignore[no-untyped-def]
        page = int(request.GET.get("page", 1))
        page_size = int(request.GET.get("page_size", 50))
        offset = (page - 1) * page_size

        qs = self.get_query_set(request=request)

        items = qs[offset : offset + page_size]
        item_serializer_class = self.get_item_serializer_class()
        count = qs.count()

        response_list = []
        for item in items:
            response_list.append(
                item_serializer_class.model_validate(item).model_dump()
            )
        return JsonResponse(
            {
                "items": response_list,
                "count": count,
                "page": page,
                "page_size": page_size,
                "has_more": count > offset + page_size,
            },
            status=200,
        )

    def get_query_set(self, request: Any) -> QuerySet:
        raise NotImplementedError("Subclasses must implement get_query_set method")

    def get_item_serializer_class(self) -> Any:
        raise NotImplementedError(
            "Subclasses must implement get_item_serializer_class method"
        )



from django.contrib import admin
from django import forms
from .models import (
    Lesson,
    Quiz,
    Course,
    ImapConnection,
    Organization,
    OrganizationUser,
    BlockedEmail,
    Enrollment,
    ContentDelivery,
    Learner,
    DeliverySchedule,
    QuizSubmission,
)


class ImapConnectionAdminForm(forms.ModelForm):
    class Meta:
        model = ImapConnection
        fields = "__all__"
        widgets = {
            "password": forms.PasswordInput(
                render_value=True,
            ),
        }


class CourseAdmin(admin.ModelAdmin):
    list_display = ("title", "enabled")
    search_fields = ("title",)
    list_filter = ("enabled",)


class ImapConnectionAdmin(admin.ModelAdmin):
    list_display = ("email", "server", "port")
    search_fields = ("email", "server")
    list_filter = ("port",)
    form = ImapConnectionAdminForm

    def get_object(self, *args, **kwargs) -> ImapConnection | None:  # type: ignore[no-untyped-def]
        obj = super().get_object(*args, **kwargs)
        if obj:
            obj.imap_password = obj.decrypt_password(obj.imap_password)
        return obj


admin.site.register(Course, CourseAdmin)
admin.site.register(ImapConnection, ImapConnectionAdmin)
admin.site.register(Organization)
admin.site.register(BlockedEmail)
admin.site.register(OrganizationUser)
admin.site.register(Enrollment)
admin.site.register(ContentDelivery)
admin.site.register(Learner)
admin.site.register(DeliverySchedule)
admin.site.register(QuizSubmission)
admin.site.register(Lesson)
admin.site.register(Quiz)



from django.utils.decorators import method_decorator
from django.views.decorators.csrf import ensure_csrf_cookie
from django.views import View
from django.views.generic.base import TemplateResponseMixin
from django.http import HttpResponse
from django.utils.translation import gettext as _
from django.utils.translation import get_language_info, get_language
from django.urls import reverse
from django_email_learning.models import ContentDelivery, EnrollmentStatus, Certificate
from django_email_learning.services import jwt_service
from django_email_learning.personalised.serializers import PublicQuizSerializer
from django_email_learning.services.command_models.verify_enrollment_command import (
    VerifyEnrollmentCommand,
)
from django_email_learning.services.command_models.unsubscribe_command import (
    UnsubscribeCommand,
)
from django.core.files.storage import default_storage
import qrcode
import uuid
import logging
import io


@method_decorator(ensure_csrf_cookie, name="dispatch")
class BaseTemplateView(View, TemplateResponseMixin):
    def error_response(
        self,
        message: str,
        exception: Exception | None,
        status_code: int = 500,
        title: str = _("Error"),
    ) -> HttpResponse:
        error_ref = uuid.uuid4().hex
        if exception:
            logging.exception(
                f"{message} - Ref: {error_ref}", extra={"error_ref": error_ref}
            )
        else:
            logging.error(
                f"{message} - Ref: {error_ref}", extra={"error_ref": error_ref}
            )
        current_lang_code = get_language()
        lang_info = get_language_info(current_lang_code)
        return self.render_to_response(
            context={
                "appContext": {
                    "ref": error_ref,
                    "errorMessage": message,
                    "direction": "rtl" if lang_info["bidi"] else "ltr",
                    "localeMessages": {
                        "error": _("Error"),
                    },
                },
                "page_title": title,
            },
            status=status_code,
        )

    def get_decoded_token(self, request) -> dict | HttpResponse:  # type: ignore[no-untyped-def]
        try:
            token = request.GET["token"]
        except KeyError as e:
            return self.error_response(
                message=_("The link is not valid."),
                exception=e,
                status_code=400,
                title=_("Invalid Link"),
            )
        try:
            return jwt_service.decode_jwt(token=token)
        except jwt_service.InvalidTokenException as e:
            return self.error_response(
                message=_("The link is not valid."),
                exception=e,
                status_code=400,
                title=_("Invalid Link"),
            )
        except jwt_service.ExpiredTokenException as e:
            return self.error_response(
                message=_("The link has expired."),
                exception=e,
                status_code=410,
                title=_("Expired Link"),
            )

    def get_app_context(self) -> dict:  # type: ignore[no-untyped-def]
        current_lang_code = get_language()
        lang_info = get_language_info(current_lang_code)
        return {
            "direction": "rtl" if lang_info["bidi"] else "ltr",
        }


class QuizPublicView(BaseTemplateView):
    template_name = "personalised/quiz_public.html"

    def get(self, request, *args, **kwargs) -> HttpResponse:  # type: ignore[no-untyped-def]
        try:
            token = request.GET["token"]
            decoded_token = self.get_decoded_token(request)
            if isinstance(decoded_token, HttpResponse):
                return decoded_token  # Return error response if token is invalid
            question_ids = decoded_token.get("question_ids", [])
            delivery = ContentDelivery.objects.get(
                id=decoded_token["delivery_id"],
                hash_value=decoded_token["delivery_hash"],
            )
            enrollment = delivery.enrollment
            if enrollment.status != EnrollmentStatus.ACTIVE:
                return self.error_response(
                    message=_("This quiz is no longer valid."),
                    exception=ValueError("Enrollment is not active"),
                    title=_("Invalid Quiz"),
                )
            quiz = delivery.course_content.quiz
            if not quiz:
                return self.error_response(
                    message=_("There is no quiz associated with this link."),
                    exception=None,
                    title=_("Invalid Quiz"),
                )
            if not delivery.course_content.is_published:
                return self.error_response(
                    message=_("There is no valid quiz associated with this link."),
                    exception=ValueError("Quiz is not published"),
                    title=_("Invalid Quiz"),
                )
            quiz_data = PublicQuizSerializer.model_validate(quiz).model_dump()
            if question_ids:
                selected_questions = [
                    q for q in quiz_data["questions"] if q["id"] in question_ids
                ]
                # Only filter questions if all specified question IDs are valid and included in the quiz
                if len(selected_questions) == len(question_ids):
                    quiz_data["questions"] = selected_questions
            return self.render_to_response(
                context={
                    "appContext": {
                        "quiz": quiz_data,
                        "token": token,
                        "csrfToken": request.META.get("CSRF_COOKIE", ""),
                        "apiEndpoint": reverse(
                            "django_email_learning:api_personalised:quiz_submission"
                        ),
                        "localeMessages": {
                            "quiz_intro": _(
                                "Please select all correct answers for each question. Note that some questions may have multiple correct answers. This quiz uses negative marking for incorrect choices; if you are unsure, it is better to leave the question unanswered."
                            ),
                            "no_answer_warning": _(
                                "You have not selected any answers for this question. Are you sure you want to proceed?"
                            ),
                            "your_score": _("Your score"),
                            "error_loading_quiz": _("Error loading quiz"),
                            "ready_to_submit": _("Ready to submit?"),
                            "submit_quiz_note": _(
                                "Please keep in mind that this quiz uses negative marking for incorrect answers. If you are unsure of an answer, it may be better to leave it blank."
                            ),
                            "cancel": _("Cancel"),
                            "submit": _("Submit"),
                            "close_window_message": _("You can now close this window!"),
                        },
                    }
                    | self.get_app_context(),
                }
            )

        except ContentDelivery.DoesNotExist as e:
            return self.error_response(
                message=_("An error occurred while retrieving the quiz"),
                exception=e,
                title=_("Error"),
                status_code=410,
            )
        except KeyError as e:
            return self.error_response(
                message=_("The link is not valid"),
                exception=e,
                status_code=400,
                title=_("Invalid Link"),
            )
        except jwt_service.InvalidTokenException as e:
            return self.error_response(
                message=_("The link is not valid"),
                exception=e,
                status_code=400,
                title=_("Invalid Link"),
            )
        except jwt_service.ExpiredTokenException as e:
            return self.error_response(
                message=_("The link has expired"),
                exception=e,
                status_code=410,
                title=_("Expired Link"),
            )


class CertificateFormView(BaseTemplateView):
    template_name = "personalised/certificate_form.html"

    def get(self, request, *args, **kwargs) -> HttpResponse:  # type: ignore[no-untyped-def]
        decoded_token = self.get_decoded_token(request)
        if isinstance(decoded_token, HttpResponse):
            return decoded_token  # Return error response if token is invalid

        return self.render_to_response(
            context={
                "appContext": {
                    "csrfToken": request.META.get("CSRF_COOKIE", ""),
                    "token": request.GET.get("token", ""),
                    "apiEndpoint": reverse(
                        "django_email_learning:api_personalised:submit_certificate_form"
                    ),
                    "localeMessages": {
                        "form_intro": _(
                            "Congratulations on completing the course! To issue your certificate, please enter the name you would like displayed on it."
                        ),
                        "full_name": _("Full Name"),
                        "full_name_required": _("Full Name is required"),
                        "error_sending_data": _(
                            "An error occurred while sending data. Please try again later."
                        ),
                        "form_submission_success": _(
                            "Your certificate name has been submitted successfully!"
                        ),
                        "submit": _("Submit"),
                        "view_certificate": _("View Certificate"),
                    },
                }
                | self.get_app_context(),
            }
        )


class VerifyEnrollmentView(BaseTemplateView):
    template_name = "personalised/command_result.html"

    def get(self, request, *args, **kwargs) -> HttpResponse:  # type: ignore[no-untyped-def]
        decoded_token = self.get_decoded_token(request)
        if isinstance(decoded_token, HttpResponse):
            return decoded_token  # Return error response if token is invalid
        enrollment_id = decoded_token["enrollment_id"]
        verification_code = decoded_token["verification_code"]

        command = VerifyEnrollmentCommand(
            command_name="verify_enrollment",
            enrollment_id=enrollment_id,
            verification_code=verification_code,
        )
        try:
            command.execute()
        except Exception as e:
            return self.error_response(
                message=_("An error occurred during enrollment verification."),
                exception=e,
                title=_("Verification Error"),
            )

        return self.render_to_response(
            context={
                "page_title": _("Enrollment Verified"),
                "appContext": {
                    "successMessage": _(
                        "Your enrollment has been successfully verified."
                    ),
                    "localeMessages": {"Confirm": _("Confirm")},
                }
                | self.get_app_context(),
            }
        )


class UnsubscribeView(BaseTemplateView):
    template_name = "personalised/command_result.html"

    def get(self, request, *args, **kwargs) -> HttpResponse:  # type: ignore[no-untyped-def]
        decoded_token = self.get_decoded_token(request)
        if isinstance(decoded_token, HttpResponse):
            return decoded_token  # Return error response if token is invalid
        if request.GET.get("confirm") != "true":
            return self.render_to_response(
                context={
                    "page_title": _("Confirm Unsubscription"),
                    "appContext": {
                        "confirmationMessage": _(
                            "Are you sure you want to unsubscribe from our mailing list?"
                        ),
                        "confirmUrl": f"{request.path}?token={request.GET.get('token')}&confirm=true",
                        "localeMessages": {"Confirm": _("Confirm")},
                    }
                    | self.get_app_context(),
                }
            )
        command = UnsubscribeCommand(
            email=decoded_token["email"],
            course_slug=decoded_token["course_slug"],
            organization_id=decoded_token["organization_id"],
        )
        try:
            command.execute()
        except Exception as e:
            return self.error_response(
                message=_("An error occurred during unsubscription."),
                exception=e,
                title=_("Unsubscription Error"),
            )
        return self.render_to_response(
            context={
                "page_title": _("Unsubscribed"),
                "appContext": {
                    "successMessage": _(
                        "You have been successfully unsubscribed from our mailing list."
                    ),
                    "localeMessages": {"Confirm": _("Confirm")},
                }
                | self.get_app_context(),
            }
        )


class CertificateView(BaseTemplateView):
    template_name = "personalised/certificate.html"

    def get(self, request, *args, **kwargs) -> HttpResponse:  # type: ignore[no-untyped-def]
        certificate_number = kwargs.get("certificate_number")
        if not certificate_number:
            return self.error_response(
                message=_("Certificate number is required."),
                exception=ValueError("Certificate number is required"),
                status_code=400,
                title=_("Invalid Certificate"),
            )
        id_parts = certificate_number.split("-")
        if len(id_parts) != 4:
            return self.error_response(
                message=_("Invalid certificate number format."),
                exception=ValueError("Invalid certificate number format"),
                status_code=400,
                title=_("Invalid Certificate"),
            )
        certificate_id = id_parts[2]
        random_suffix = id_parts[3]
        try:
            certificate = Certificate.objects.get(
                id=certificate_id, random_suffix=random_suffix
            )
        except Certificate.DoesNotExist:
            return self.error_response(
                message=_("Certificate not found."),
                exception=ValueError("Certificate not found"),
                status_code=404,
                title=_("Certificate Not Found"),
            )

        path = reverse(
            "django_email_learning:personalised:certificate",
            kwargs={"certificate_number": certificate_number},
        )
        full_url = request.build_absolute_uri(path)

        qr = qrcode.QRCode(version=1, box_size=10, border=4)
        qr.add_data(full_url)
        qr.make(fit=True)

        qrcode_img = qr.make_image(
            fill_color="black", back_color="transparent"
        ).convert("RGBA")  # type: ignore[union-attr]

        img_byte_arr = io.BytesIO()
        qrcode_img.save(img_byte_arr, format="PNG")

        media_path = f"certificates/{certificate.enrollment.course.organization.id}/{certificate_number}.png"
        default_storage.save(media_path, img_byte_arr)
        absolute_media_url = default_storage.url(media_path)

        return self.render_to_response(
            context={
                "page_title": f"{_("Certificate of Completion")} | {certificate.enrollment.course.title} | {certificate.name_on_certificate}",
                "appContext": {
                    "name": certificate.name_on_certificate,
                    "courseTitle": certificate.enrollment.course.title,
                    "issueDate": certificate.issued_at.strftime("%B %d, %Y"),
                    "certificateNumber": certificate_number,
                    "organizationName": certificate.enrollment.course.organization.name,
                    "logoUrl": certificate.enrollment.course.organization.logo.url
                    if certificate.enrollment.course.organization.logo
                    else "",
                    "qrcodeUrl": absolute_media_url,
                    "localeMessages": {
                        "title": _("Certificate of Completion"),
                        "description": _(
                            "This certifies that {name} has successfully completed the {course_title} course"
                        ).format(
                            name=certificate.name_on_certificate,
                            course_title=certificate.enrollment.course.title,
                        ),
                        "issue_date": _("Issued on"),
                        "certificate_number": _("Certificate Number"),
                        "organization_team": _("{organization_name} Team").format(
                            organization_name=certificate.enrollment.course.organization.name
                        ),
                    },
                }
                | self.get_app_context(),
            }
        )



from django.views import View
from django.http import JsonResponse
from pydantic import ValidationError
from django_email_learning.public.api.serializers import EnrollmentRequest
from django_email_learning.services.command_models.enroll_command import EnrollCommand
from django_email_learning.services.command_models.exceptions.invalid_course_slug_error import (
    InvalidCourseSlugError,
)
from django_email_learning.services.command_models.exceptions.blocked_email_error import (
    BlockedEmailError,
)
from django_email_learning.services.command_models.exceptions.enrollment_already_exists_error import (
    EnrollmentAlreadyExistsError,
)
import json
import logging

logger = logging.getLogger(__name__)


class EnrollView(View):
    def post(self, request, *args, **kwargs) -> JsonResponse:  # type: ignore[no-untyped-def]
        # Logic for enrolling a user via public API
        payload = json.loads(request.body)
        try:
            serlizer = EnrollmentRequest.model_validate(payload)
            command = EnrollCommand(
                email=serlizer.email,
                course_slug=serlizer.course_slug,
                organization_id=serlizer.organization_id,
            )

            try:
                command.execute()
                return JsonResponse({"status": "enrolled"}, status=200)
            except EnrollmentAlreadyExistsError as e:
                logger.info(f"Enrollment already exists: {e}")
                return JsonResponse({"status": "already_enrolled"}, status=200)
            except BlockedEmailError as e:
                logger.error(f"Blocked email error: {e}")
                return JsonResponse({"error": str(e)}, status=403)
            except InvalidCourseSlugError as e:
                logger.error(f"Invalid course slug error: {e}")
                return JsonResponse({"error": str(e)}, status=400)

        except ValidationError as e:
            return JsonResponse({"error": str(e)}, status=400)



from pydantic import BaseModel, field_serializer


class PublicCourseSerializer(BaseModel):
    id: int
    title: str
    slug: str
    description: str | None = None
    imap_email: str | None = None
    image: str | None = None
    language: str
    is_rtl: bool = False
    lessons: list[str] = []

    @field_serializer("description")
    def serialize_description_with_br(self, description: str | None) -> str | None:
        if description is not None:
            return description.replace("\n", "<br />")
        return description

    @field_serializer("language")
    def serialize_language(self, language_code: str) -> str:
        from django.conf.global_settings import LANGUAGES

        language_dict = dict(LANGUAGES)
        return language_dict.get(language_code, language_code)


class OrganizationSerializer(BaseModel):
    id: int
    name: str
    logo_url: str | None = None
    description: str | None = None
    courses: list[PublicCourseSerializer] = []
    public_url: str

    @field_serializer("description")
    def serialize_description_with_br(self, description: str | None) -> str | None:
        if description is not None:
            return description.replace("\n", "<br>")
        return description



from django.urls import path
from django_email_learning.jobs.api.views import DeliverContentsJobView

app_name = "django_email_learning"

urlpatterns = [
    path(
        "deliver_contents/", DeliverContentsJobView.as_view(), name="deliver_contents"
    ),
]



import base64
import ipaddress
import re
import random
import uuid
import logging
from enum import StrEnum
from typing import Any
from django.conf import settings
from django.conf.global_settings import LANGUAGES
from django.core.files.storage import default_storage
from django.urls import reverse
from django.db import models, transaction
from django.core.validators import (
    MaxValueValidator,
    MinValueValidator,
    MinLengthValidator,
)
from django_email_learning.services.email_sender_service import EmailSenderService
from django_email_learning.services.metrics_service import MetricsService
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string
from django.core.exceptions import ImproperlyConfigured
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives import hashes
from django.forms import ValidationError
from django.contrib.auth.models import User
from django.utils import timezone
from django.utils.translation import gettext as _
from django.utils.translation import ngettext
from datetime import timedelta
from django_email_learning.services import jwt_service

from PIL import Image
from typing import Optional
from datetime import datetime


logger = logging.getLogger(__name__)


class EnrollmentStatus(StrEnum):
    UNVERIFIED = "unverified"
    ACTIVE = "active"
    COMPLETED = "completed"
    DEACTIVATED = "deactivated"


class DeactivationReason(StrEnum):
    CANCELED = "canceled"
    BLOCKED = "blocked"
    FAILED = "failed"
    INACTIVE = "inactive"


class DeliveryStatus(StrEnum):
    SCHEDULED = "scheduled"
    PROCESSING = "processing"
    DELIVERED = "delivered"
    CANCELED = "canceled"
    BLOCKED = "blocked"


METRIC_SERVICE = MetricsService()


def is_domain_or_ip(value: str) -> None:
    """
    Validate if the given value is a valid domain name or IP address.

    Raises:
        ValueError: If the value is not a valid domain or IP address.
    """
    try:
        ipaddress.ip_address(value)
    except ValueError:
        DOMAIN_REGEX = re.compile(r"^(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z]{2,}$")
        if not bool(DOMAIN_REGEX.match(value.lower())):
            raise ValueError(f"{value} is not a valid domain or IP address")


class Organization(models.Model):
    name = models.CharField(max_length=200, unique=True)
    logo = models.ImageField(upload_to="organization_logos/", null=True, blank=True)
    description = models.TextField(null=True, blank=True)

    def __str__(self) -> str:
        return self.name

    @property
    def public_url(self) -> str:
        path = reverse(
            "django_email_learning:public:organization_view",
            kwargs={"organization_id": self.id},
        )
        return f"{settings.DJANGO_EMAIL_LEARNING['SITE_BASE_URL']}{path}"

    def replace_logo(self, file_path: str) -> str:
        if default_storage.exists(file_path):
            allowed_extensions = [".jpg", ".jpeg", ".png", ".svg"]
            if not any(file_path.lower().endswith(ext) for ext in allowed_extensions):
                raise ValueError("Logo must be an image file with a valid extension.")
            final_path = f"organization_logos/{self.id}/{file_path.split('/')[-1]}"
            default_storage.save(final_path, default_storage.open(file_path))
            self.logo = final_path
            self.save()
            return final_path
        else:
            raise ValueError("Logo file does not exist.")


class OrganizationUser(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE, related_name="memberships")
    organization = models.ForeignKey(Organization, on_delete=models.CASCADE)
    role = models.CharField(
        max_length=50,
        choices=[
            ("admin", "Admin"),
            ("editor", "Editor"),
            ("viewer", "Viewer"),
        ],
        db_index=True,
    )

    def __str__(self) -> str:
        return f"{self.user.username} - {self.organization.name}"

    class Meta:
        unique_together = [["user", "organization"]]


class EncryptionMixin(models.Model):
    salt = models.CharField(max_length=32, editable=False)

    @classmethod
    def _fernet(cls, salt: str) -> Fernet:
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt.encode(),
            iterations=100000,
        )
        DJANGO_EMAIL_LEARNING_SETTINGS: dict = getattr(
            settings, "DJANGO_EMAIL_LEARNING", {}
        )
        try:
            secret = DJANGO_EMAIL_LEARNING_SETTINGS["ENCRYPTION_SECRET_KEY"]
        except KeyError:
            raise ImproperlyConfigured(
                "DJANGO_EMAIL_LEARNING['ENCRYPTION_SECRET_KEY'] must be set in settings.py"
            )
        key = base64.urlsafe_b64encode(kdf.derive(secret.encode()))
        return Fernet(key)

    @classmethod
    def encrypted_value(cls, value: str, salt: str) -> str:
        f = cls._fernet(salt)
        return f.encrypt(value.encode()).decode()

    def _encrypt_password(self, password: str) -> str:
        if not self.salt:
            self.salt = (
                base64.urlsafe_b64encode(uuid.uuid4().bytes).decode().rstrip("=")
            )
        f = self._fernet(self.salt)
        return f.encrypt(password.encode()).decode()

    def decrypt_password(self, encrypted_password: str) -> str:
        f = self._fernet(self.salt)
        return f.decrypt(encrypted_password.encode()).decode()

    class Meta:
        abstract = True


class ImapConnection(EncryptionMixin):
    server = models.CharField(max_length=200, validators=[is_domain_or_ip])
    port = models.IntegerField(db_default=993)
    email = models.EmailField(max_length=200, unique=True)
    password = models.CharField(max_length=200)
    organization = models.ForeignKey(Organization, on_delete=models.CASCADE)

    def __str__(self) -> str:
        return f"{self.email}|{self.server}:{self.port}"

    def save(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        if self.password:
            try:
                self.decrypt_password(self.password)
                # Password is already encrypted
            except InvalidToken:
                self.password = self._encrypt_password(self.password)
        if self.server:
            self.server = self.server.lower()
        self.full_clean()
        super().save(*args, **kwargs)


class Course(models.Model):
    title = models.CharField(max_length=200)
    slug = models.SlugField(
        max_length=50,
        help_text="A short label for the course, used in URLs or email interactive actions. You can not edit it later.",
    )
    description = models.TextField(null=True, blank=True)
    enabled = models.BooleanField(default=False)
    imap_connection = models.ForeignKey(
        ImapConnection, on_delete=models.SET_NULL, null=True, blank=True
    )
    organization = models.ForeignKey(Organization, on_delete=models.CASCADE)
    image = models.ImageField(upload_to="course_images/", null=True, blank=True)
    language = models.CharField(
        max_length=10,
        choices=LANGUAGES,
        default="en",
    )

    def __str__(self) -> str:
        return self.title

    class Meta:
        unique_together = [["slug", "organization"], ["title", "organization"]]

    def delete(
        self, using: Any | None = None, keep_parents: bool = False
    ) -> tuple[int, dict[str, int]]:
        if self.enabled:
            raise ValueError(
                "Course can not be deleted when enabled, please disable the course first!"
            )
        return super().delete(using, keep_parents)

    @property
    def enrollments_count(self) -> dict[str, int]:
        unverified_count = self.enrollment_set.filter(
            status=EnrollmentStatus.UNVERIFIED
        ).count()
        active_count = self.enrollment_set.filter(
            status=EnrollmentStatus.ACTIVE
        ).count()
        completed_count = self.enrollment_set.filter(
            status=EnrollmentStatus.COMPLETED
        ).count()
        deactivated_count = self.enrollment_set.filter(
            status=EnrollmentStatus.DEACTIVATED
        ).count()
        total_count = self.enrollment_set.count()
        return {
            EnrollmentStatus.UNVERIFIED: unverified_count,
            EnrollmentStatus.ACTIVE: active_count,
            EnrollmentStatus.COMPLETED: completed_count,
            EnrollmentStatus.DEACTIVATED: deactivated_count,
            "total": total_count,
        }

    def generate_unsubscribe_link(self, email: str) -> str:
        payload = {
            "email": email,
            "course_slug": self.slug,
            "organization_id": self.organization.id,
        }
        token = jwt_service.generate_jwt(payload=payload)
        unsubscribe_path = reverse("django_email_learning:personalised:unsubscribe")
        link = f"{settings.DJANGO_EMAIL_LEARNING['SITE_BASE_URL']}{unsubscribe_path}?token={token}"
        return link

    def replace_image(self, file_path: str) -> str:
        if default_storage.exists(file_path):
            with default_storage.open(file_path) as f:
                img = Image.open(f)
                width, height = img.size
                if width < 580 or height < 360:
                    raise ValueError(
                        "Image dimensions must be at least 580x360 pixels."
                    )
            allowed_extensions = [".jpg", ".jpeg", ".png", ".svg"]
            if not any(file_path.lower().endswith(ext) for ext in allowed_extensions):
                raise ValueError("Image must be an image file with a valid extension.")
            final_path = f"organization/{self.organization.id}/course_images/{self.id}_{file_path.split('/')[-1]}"
            default_storage.save(final_path, default_storage.open(file_path))
            self.image = final_path
            self.save()
            return final_path
        else:
            raise ValueError("Image file does not exist.")


class Lesson(models.Model):
    title = models.CharField(max_length=200)
    content = models.TextField()

    def __str__(self) -> str:
        return self.title


class QuizSelectionStrategy(StrEnum):
    ALL_QUESTIONS = "all"
    RANDOM_QUESTIONS = "random"


class Quiz(models.Model):
    title = models.CharField(max_length=500)
    required_score = models.IntegerField(validators=[MaxValueValidator(100)])
    selection_strategy = models.CharField(
        max_length=50,
        choices=[
            (QuizSelectionStrategy.ALL_QUESTIONS.value, "All Questions"),
            (QuizSelectionStrategy.RANDOM_QUESTIONS.value, "Random Questions"),
        ],
    )
    deadline_days = models.IntegerField(
        help_text="Time limit to complete the quiz in days. Minimum is 1 day and maximum is 30 days.",
        validators=[MinValueValidator(1), MaxValueValidator(30)],
    )

    class Meta:
        verbose_name_plural = "Quizzes"

    def __str__(self) -> str:
        return self.title

    def validate_questions(self) -> None:
        if not self.questions.exists():
            raise ValidationError("At least one question is required.")

        for question in self.questions.all():
            try:
                question.validate_answers()
            except ValidationError as e:
                raise ValidationError(f"For question '{question.text}', {e.message}")

    def random_question_ids(self) -> list[int]:
        question_ids = list(self.questions.values_list("id", flat=True))
        if self.selection_strategy == QuizSelectionStrategy.ALL_QUESTIONS.value:
            return question_ids
        if len(question_ids) <= 5:
            return question_ids
        number_of_questions = int(max(5, len(question_ids) // 1.5))
        selected_ids = random.sample(question_ids, k=number_of_questions)
        return selected_ids


class Question(models.Model):
    quiz = models.ForeignKey(Quiz, on_delete=models.CASCADE, related_name="questions")
    text = models.CharField(max_length=500)
    priority = models.IntegerField()

    def __str__(self) -> str:
        return self.text

    def validate_answers(self) -> None:
        if not self.answers.filter(is_correct=True).exists():
            raise ValueError("At least one correct answer is required.")

        if self.answers.count() < 2:
            raise ValueError("At least two answers are required.")

    def is_multiple_choice(self) -> bool:
        return self.answers.filter(is_correct=True).count() > 1


class Answer(models.Model):
    question = models.ForeignKey(
        Question, on_delete=models.CASCADE, related_name="answers"
    )
    text = models.CharField(max_length=500)
    is_correct = models.BooleanField(default=False)

    def __str__(self) -> str:
        return self.text

    def delete(self, *args, **kwargs) -> tuple[int, dict[str, int]]:  # type: ignore[no-untyped-def]
        if self.question.quiz.coursecontent_set.filter(is_published=True).exists():
            raise ValidationError("Cannot delete answers from a published quiz.")
        return super().delete(*args, **kwargs)


class CourseContent(models.Model):
    course = models.ForeignKey(Course, on_delete=models.CASCADE)
    priority = models.IntegerField()
    type = models.CharField(
        max_length=50,
        choices=[
            ("lesson", "Lesson"),
            ("quiz", "Quiz"),
        ],
    )
    lesson = models.ForeignKey(Lesson, null=True, blank=True, on_delete=models.CASCADE)
    quiz = models.ForeignKey(Quiz, null=True, blank=True, on_delete=models.CASCADE)
    waiting_period = models.IntegerField(
        help_text="Waiting period in seconds after previous content is sent or submited."
    )
    is_published = models.BooleanField(default=False)

    def __str__(self) -> str:
        if self.type == "lesson" and self.lesson:
            return f"{self.priority} - Lesson: {self.lesson.title}"
        elif self.type == "quiz" and self.quiz:
            return f"{self.priority} - Quiz: {self.quiz.title}"
        return f"{self.course.title} content #{self.priority}"

    @property
    def title(self) -> str:
        if self.type == "lesson" and self.lesson:
            return self.lesson.title
        elif self.type == "quiz" and self.quiz:
            return self.quiz.title
        return "Untitled Content"

    def human_readable_waiting_period(self) -> str:
        if self.waiting_period < 60:
            return ngettext(
                "%(count)d second", "%(count)d seconds", self.waiting_period
            ) % {"count": self.waiting_period}
        elif self.waiting_period < 3600:
            minutes = self.waiting_period // 60
            return ngettext("%(count)d minute", "%(count)d minutes", minutes) % {
                "count": minutes
            }
        elif self.waiting_period < 86400:
            hours = self.waiting_period // 3600
            return ngettext("%(count)d hour", "%(count)d hours", hours) % {
                "count": hours
            }
        else:
            days = self.waiting_period // 86400
            return ngettext("%(count)d day", "%(count)d days", days) % {"count": days}

    def _validate_content(self) -> None:
        if self.type == "lesson" and not self.lesson:
            raise ValidationError("Lesson must be provided for lesson content.")
        if self.type == "quiz" and not self.quiz:
            raise ValidationError("Quiz must be provided for quiz content.")
        if self.type == "lesson" and self.lesson:
            self.lesson.full_clean()
        elif self.type == "quiz" and self.quiz:
            self.quiz.full_clean()

    def full_clean(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        self._validate_content()
        return super().full_clean(*args, **kwargs)

    def save(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        self.full_clean()
        super().save(*args, **kwargs)

    def get_next(self) -> Optional["CourseContent"]:
        next_content = (
            CourseContent.objects.filter(
                course=self.course, is_published=True, priority__gt=self.priority
            )
            .order_by("priority")
            .first()
        )
        return next_content

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["course", "quiz"],
                condition=models.Q(quiz__isnull=False),
                name="unique_quiz_per_course",
            ),
            models.UniqueConstraint(
                fields=["course", "lesson"],
                condition=models.Q(lesson__isnull=False),
                name="unique_lesson_per_course",
            ),
            models.UniqueConstraint(
                fields=["course", "priority"],
                name="unique_priority_per_course",
            ),
        ]


class BlockedEmail(models.Model):
    email = models.EmailField(unique=True)

    def __str__(self) -> str:
        return self.email

    def save(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        self.email = self.email.lower()
        self.full_clean()
        super().save(*args, **kwargs)


class Learner(models.Model):
    organization = models.ForeignKey(Organization, on_delete=models.CASCADE)
    email = models.EmailField()
    created_at = models.DateTimeField(auto_now_add=True)

    def save(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        self.email = self.email.lower()
        self.full_clean()
        super().save(*args, **kwargs)

    @property
    def enrollments_count(self) -> dict[str, int]:
        return {
            "total": self.enrollment_set.count(),
            "completed": self.enrollment_set.filter(
                status=EnrollmentStatus.COMPLETED
            ).count(),
        }

    class Meta:
        unique_together = [["organization", "email"]]


class Enrollment(models.Model):
    state_transitions = {
        EnrollmentStatus.UNVERIFIED: [
            EnrollmentStatus.ACTIVE,
            EnrollmentStatus.DEACTIVATED,
        ],
        EnrollmentStatus.ACTIVE: [
            EnrollmentStatus.COMPLETED,
            EnrollmentStatus.DEACTIVATED,
        ],
        EnrollmentStatus.COMPLETED: [],
        EnrollmentStatus.DEACTIVATED: [],
    }
    learner = models.ForeignKey(Learner, on_delete=models.CASCADE)
    course = models.ForeignKey(Course, on_delete=models.CASCADE)
    enrolled_at = models.DateTimeField(auto_now_add=True)
    activated_at = models.DateTimeField(null=True, blank=True)
    final_state_at = models.DateTimeField(null=True, blank=True)
    status = models.CharField(
        max_length=50,
        choices=[
            (EnrollmentStatus.UNVERIFIED, "Unverified"),
            (EnrollmentStatus.ACTIVE, "Active"),
            (EnrollmentStatus.COMPLETED, "Completed"),
            (EnrollmentStatus.DEACTIVATED, "Deactivated"),
        ],
        default=EnrollmentStatus.UNVERIFIED,
    )
    deactivation_reason = models.CharField(
        null=True,
        blank=True,
        choices=[
            (DeactivationReason.CANCELED, "Canceled"),
            (DeactivationReason.BLOCKED, "Blocked"),
            (DeactivationReason.FAILED, "Failed"),
            (DeactivationReason.INACTIVE, "Inactive"),
        ],
        max_length=50,
    )
    activation_code = models.CharField(max_length=6, null=True, blank=True)

    def save(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        if self.pk:
            old_status = Enrollment.objects.get(pk=self.pk).status
            old_status = EnrollmentStatus(old_status)
            if old_status != self.status:
                allowed_transitions = self.state_transitions.get(old_status, [])
                if self.status not in allowed_transitions:
                    raise ValidationError(
                        f"Invalid status transition from {old_status} to {self.status}."
                    )
        else:
            self.activation_code = "".join(random.choices("0123456789", k=6))
        if self.status != "deactivated" and self.deactivation_reason is not None:
            raise ValidationError(
                "Deactivation reason must be null unless status is 'deactivated'."
            )
        if self.status == "deactivated" and not self.deactivation_reason:
            raise ValidationError(
                "Deactivation reason must be provided when status is 'deactivated'."
            )
        self.full_clean()
        if self.status == EnrollmentStatus.ACTIVE and self.activated_at is None:
            self.activated_at = timezone.now()
        if self.status in [EnrollmentStatus.COMPLETED, EnrollmentStatus.DEACTIVATED]:
            if self.final_state_at is None:
                self.final_state_at = timezone.now()

        super().save(*args, **kwargs)

    def __str__(self) -> str:
        return f"{self.learner.email} - {self.course.title} ({self.status})"

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["learner", "course"],
                condition=models.Q(status__in=["unverified", "active", "completed"]),
                name="unique_active_enrollment",
            )
        ]

    def graduate(self) -> None:
        with transaction.atomic():
            if self.status != EnrollmentStatus.ACTIVE:
                raise ValidationError(
                    "Only active enrollments can be marked as completed."
                )
            self.status = EnrollmentStatus.COMPLETED
            self.final_state_at = timezone.now()
            METRIC_SERVICE.user_completed_course(
                course_slug=self.course.slug,
                organization_id=self.course.organization.id,
            )
            logger.info(
                f"Learner ID {self.learner.id} has completed the course {self.course.title}."
            )
            self.save()
            self.send_certificate_form()

    def send_certificate_form(self) -> None:
        if self.status != EnrollmentStatus.COMPLETED:
            raise ValidationError(
                "Certificate form can only be sent for completed enrollments."
            )
        token_payload = {
            "enrollment_id": self.id,
        }
        logging.info(
            f"Executing SendCertificateFormCommand for enrollment ID {self.id}"
        )
        token = jwt_service.generate_jwt(token_payload, exp=datetime.max)
        certificate_path = reverse(
            "django_email_learning:personalised:certificate_form"
        )
        link = f"{settings.DJANGO_EMAIL_LEARNING['SITE_BASE_URL']}{certificate_path}?token={token}"

        subject = _("Finalize your Certificate")

        context = {
            "course_title": self.course.title,
            "organization_name": self.course.organization.name,
            "link": link,
        }
        payload = render_to_string("emails/certificate_form.txt", context)

        email_service = EmailSenderService()
        email_message = EmailMultiAlternatives(
            subject=subject,
            body=payload,
            from_email=email_service.from_email,
            to=[self.learner.email],
        )
        email_message.attach_alternative(
            render_to_string("emails/certificate_form.html", context), "text/html"
        )

        email_service.send(email_message)
        logging.info(f"Certificate form email sent for enrollment ID {self.id}")

    def fail(self) -> None:
        if self.status != EnrollmentStatus.ACTIVE:
            raise ValidationError("Only active enrollments can be marked as failed.")
        self.status = EnrollmentStatus.DEACTIVATED
        self.deactivation_reason = DeactivationReason.FAILED
        self.final_state_at = timezone.now()
        METRIC_SERVICE.user_enrollment_deactivated(
            course_slug=self.course.slug,
            organization_id=self.course.organization.id,
            reason=DeactivationReason.FAILED,
        )
        logger.info(
            f"Learner ID {self.learner.id} has failed the course {self.course.title}."
        )
        self.save()

    @transaction.atomic()
    def schedule_first_content_delivery(self) -> None:
        first_content = (
            CourseContent.objects.filter(course=self.course, is_published=True)
            .order_by("priority")
            .first()
        )
        if first_content:
            delivery = ContentDelivery.objects.create(
                enrollment=self,
                course_content=first_content,
            )
            scheduled = DeliverySchedule.objects.create(
                time=timezone.now() + timedelta(seconds=first_content.waiting_period),
                delivery=delivery,
            )
            scheduled.generate_link()
        else:
            raise ValidationError("No published content available to schedule.")

    @property
    def progress_percentage(self) -> int:
        total_content = self.course.coursecontent_set.filter(is_published=True).count()
        if total_content == 0:
            return 0
        delivered_content = (
            ContentDelivery.objects.filter(
                enrollment=self,
                delivery_schedules__status=DeliveryStatus.DELIVERED,
                course_content__is_published=True,
            )
            .distinct()
            .count()
        )
        progress = int((delivered_content / total_content) * 100)
        return progress


class Certificate(models.Model):
    enrollment = models.OneToOneField(
        Enrollment, on_delete=models.CASCADE, related_name="certificate"
    )
    issued_at = models.DateTimeField(auto_now_add=True)
    name_on_certificate = models.CharField(max_length=200)
    random_suffix = models.IntegerField()

    @property
    def certificate_number(self) -> str:
        return f"{self.enrollment.course.id}-{self.enrollment.id}-{self.id}-{self.random_suffix}"

    def save(  # type: ignore[no-untyped-def]
        self, *, force_insert=False, force_update=False, using=None, update_fields=None
    ):
        if not self.random_suffix:
            self.random_suffix = random.randint(100000, 999999)
        return super().save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
        )


class ContentDelivery(models.Model):
    enrollment = models.ForeignKey(
        Enrollment, on_delete=models.CASCADE, related_name="content_deliveries"
    )
    course_content = models.ForeignKey(CourseContent, on_delete=models.CASCADE)
    hash_value = models.CharField(max_length=64, null=True, blank=True)
    valid_until = models.DateTimeField(null=True, blank=True)

    class Meta:
        unique_together = [["enrollment", "course_content"]]

    @property
    def times_delivered(self) -> int:
        return self.delivery_schedules.filter(status=DeliveryStatus.DELIVERED).count()  # type: ignore[misc]

    def update_hash(self) -> None:
        self.hash_value = (
            base64.urlsafe_b64encode(uuid.uuid4().bytes).decode().rstrip("=")
        )
        self.save()

    def schedule_next_delivery(self) -> Optional["ContentDelivery"]:
        """
        Schedules the next content delivery based on the current content's priority.
        Returns the ID of the newly created ContentDelivery if successful, otherwise None.
        """

        next_content = (
            CourseContent.objects.filter(
                course=self.course_content.course,
                is_published=True,
                priority__gt=self.course_content.priority,
            )
            .order_by("priority")
            .first()
        )
        if next_content:
            delivery, created = ContentDelivery.objects.get_or_create(
                enrollment=self.enrollment,
                course_content=next_content,
            )
            schedule = DeliverySchedule.objects.create(
                time=timezone.now() + timedelta(seconds=next_content.waiting_period),
                delivery=delivery,
            )
            schedule.generate_link()
            return delivery
        return None

    def repeat_delivery_in_days(self, days: int) -> bool:
        """
        Schedules a repeat delivery of the current content after a specified number of days.
        Returns True if the repeat delivery was scheduled, otherwise False.
        """
        schedule = DeliverySchedule.objects.create(
            time=timezone.now() + timedelta(days=days),
            delivery=self,
        )
        schedule.generate_link()
        logger.info(
            f"Repeat delivery scheduled for ContentDelivery ID {self.id} in {days} days."
        )
        return True

    def save(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        self.full_clean()
        if not self.hash_value:
            self.hash_value = (
                base64.urlsafe_b64encode(uuid.uuid4().bytes).decode().rstrip("=")
            )
        if self.course_content.quiz and not self.valid_until:
            self.valid_until = timezone.now() + timedelta(
                days=self.course_content.quiz.deadline_days
            )
        super().save(*args, **kwargs)

    def __str__(self) -> str:
        return f"Delivery of {self.course_content.title} to {self.enrollment.learner.email}"


class DeliverySchedule(models.Model):
    delivery = models.ForeignKey(
        ContentDelivery, on_delete=models.CASCADE, related_name="delivery_schedules"
    )
    time = models.DateTimeField(default=timezone.now, db_index=True)
    link = models.URLField(null=True, blank=True, max_length=500)
    status = models.CharField(
        max_length=50,
        choices=[
            (DeliveryStatus.SCHEDULED, "Scheduled"),
            (DeliveryStatus.PROCESSING, "Processing"),
            (DeliveryStatus.DELIVERED, "Delivered"),
            (DeliveryStatus.CANCELED, "Canceled"),
            (DeliveryStatus.BLOCKED, "Blocked"),
        ],
        default=DeliveryStatus.SCHEDULED,
        db_index=True,
    )
    failed_attempts = models.IntegerField(default=0)
    delivered_at = models.DateTimeField(null=True, blank=True)

    def generate_link(self) -> str:
        payload = {
            "delivery_id": self.delivery.id,
            "delivery_hash": self.delivery.hash_value,
        }

        if self.delivery.course_content.quiz:
            if (
                self.delivery.course_content.quiz.selection_strategy
                == QuizSelectionStrategy.RANDOM_QUESTIONS.value
            ):
                payload[
                    "question_ids"
                ] = self.delivery.course_content.quiz.random_question_ids()  # type: ignore[assignment]
            exp = self.time + timedelta(
                days=self.delivery.course_content.quiz.deadline_days
            )
            token = jwt_service.generate_jwt(payload=payload, exp=exp)
            quiz_path = reverse("django_email_learning:personalised:quiz_public_view")
            link = f"{settings.DJANGO_EMAIL_LEARNING['SITE_BASE_URL']}{quiz_path}?token={token}"
            self.link = link
            self.save()
            return link
        else:
            # TODO: Implement lesson link generation
            return ""

    def __str__(self) -> str:
        return f"Delivery for {self.delivery.course_content.title} to {self.delivery.enrollment.learner.email} at {self.time} - Status: {self.status}"

    def save(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        if self.status == DeliveryStatus.DELIVERED and not self.delivered_at:
            self.delivered_at = timezone.now()
        super().save(*args, **kwargs)


class QuizSubmission(models.Model):
    delivery = models.ForeignKey(
        ContentDelivery, on_delete=models.CASCADE, related_name="quiz_submissions"
    )
    score = models.IntegerField()
    is_passed = models.BooleanField()
    submitted_at = models.DateTimeField(auto_now_add=True)

    def save(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        if self.delivery.course_content.type != "quiz":
            raise ValidationError("Sent item must be associated with a quiz content.")
        already_submitted = QuizSubmission.objects.filter(
            delivery=self.delivery
        ).count()
        if already_submitted >= self.delivery.times_delivered:
            raise ValidationError(
                "Quiz submission count exceeds the number of times the quiz was sent."
            )
        self.full_clean()
        super().save(*args, **kwargs)


class ApiKey(EncryptionMixin):
    key = models.CharField(
        max_length=256, unique=True, validators=[MinLengthValidator(50)]
    )
    created_at = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(
        User, on_delete=models.SET_NULL, null=True, blank=True
    )

    @classmethod
    def generate_key(cls) -> str:
        return (
            base64.urlsafe_b64encode(uuid.uuid4().bytes + uuid.uuid4().bytes)
            .decode()
            .rstrip("=")
        )

    def save(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        try:
            self.decrypt_password(self.key)
            # Key is already encrypted
        except InvalidToken:
            self.key = self._encrypt_password(self.key)
        self.full_clean()
        super().save(*args, **kwargs)


class JobName(StrEnum):
    DELIVER_CONTENTS = "deliver_contents"


class JobStatus(StrEnum):
    RUNNING = "running"
    COMPLETED = "completed"


class JobExecution(models.Model):
    job_name = models.CharField(
        max_length=200, choices=[(job.name, job.value) for job in JobName]
    )
    started_at = models.DateTimeField(auto_now_add=True)
    finished_at = models.DateTimeField(null=True, blank=True)
    status = models.CharField(
        max_length=50,
        choices=[(status.name, status.value) for status in JobStatus],
    )

    def __str__(self) -> str:
        return (
            f"Job: {self.job_name} started at {self.started_at} - Status: {self.status}"
        )



from django.views.generic import TemplateView
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import ensure_csrf_cookie
from django.middleware.csrf import get_token
from django.db.models import Prefetch
from django_email_learning.models import Organization, Course
from django.utils.translation import get_language_info, get_language
from django.http import Http404
from django.urls import reverse
from django.utils.translation import gettext as _
from django.conf import settings
from django_email_learning.public.serializers import (
    OrganizationSerializer,
    PublicCourseSerializer,
)


def json_ld_from(
    courses: list[PublicCourseSerializer], organization: Organization
) -> dict:
    course_list = []
    for course in courses:
        course_data = {
            "@type": "Course",
            "name": course.title,
            "description": course.description,
            "inLanguage": course.language,
            "provider": {
                "@type": "Organization",
                "name": organization.name,
            },
        }
        if course.image:
            course_data["image"] = course.image
        course_list.append(course_data)
    return {
        "@context": "https://schema.org",
        "@type": "ItemList",
        "itemListElement": course_list,
    }


@method_decorator(ensure_csrf_cookie, name="dispatch")
class OrganizationView(TemplateView):
    template_name = "public/organization.html"

    def get_context_data(self, **kwargs) -> dict:  # type: ignore[no-untyped-def]
        get_token(self.request)  # Ensure CSRF token is set in cookies
        organization_id: int = kwargs.get("organization_id")  # type: ignore[assignment]
        context = super().get_context_data(**kwargs)
        # Add any additional context if needed
        organization_details = Organization.objects.filter(
            id=organization_id
        ).prefetch_related(
            Prefetch(
                "course_set",
                queryset=Course.objects.filter(enabled=True),
                to_attr="courses",
            ),
        )
        if organization_details.exists():
            organization = organization_details.first()
            if not organization:
                raise Http404(_("Organization does not exist"))
            courses = []
            for course in organization.courses:
                course_lang_info = get_language_info(course.language)
                course_data = PublicCourseSerializer(
                    id=course.id,
                    title=course.title,
                    slug=course.slug,
                    description=course.description,
                    image=self.request.build_absolute_uri(course.image.url)
                    if course.image
                    else None,
                    imap_email=course.imap_connection.email
                    if course.imap_connection
                    else None,
                    language=course.language,
                    is_rtl=course_lang_info["bidi"],
                )
                courses.append(course_data)
            organization_data = OrganizationSerializer(
                id=organization.id,
                name=organization.name,
                logo_url=organization.logo.url if organization.logo else None,
                description=organization.description,
                courses=courses,
                public_url=organization.public_url,
            )
            enroll_api_path = reverse("django_email_learning:api_public:enroll")
            current_lang_code = get_language()
            lang_info = get_language_info(current_lang_code)
            context["appContext"] = {
                "organization": organization_data.model_dump(),
                "enrollApiUrl": f"{settings.DJANGO_EMAIL_LEARNING['SITE_BASE_URL']}{enroll_api_path}",
                "direction": "rtl" if lang_info["bidi"] else "ltr",
                "localeMessages": {
                    "courses": _("Courses"),
                    "enroll_now": _("Enroll Now"),
                    "enrol_for_course": _("Enroll for COURSE_NAME"),
                    "email": _("email"),
                    "cancel": _("Cancel"),
                    "submit": _("Submit"),
                    "enrollment_success": _("You are enrolled in this course."),
                    "enrollment_failed": _("Enrollment failed. Please try again."),
                    "no_courses_available": _("No courses available."),
                    "email_required": _("Email is required"),
                    "email_invalid": _("Please enter a valid email address"),
                    "course_language": _("Course language"),
                    "in_app_browser_or_disabled_cookies": _(
                        "It seems you are using an in-app browser or have disabled cookies. Please open this link in a regular browser and ensure cookies are enabled to enroll in courses."
                    ),
                    "continue": _("Continue"),
                },
            }
            context["organization_name"] = organization.name
            context["organization_description"] = organization.description
            context["organization_logo_url"] = (
                self.request.build_absolute_uri(organization.logo.url)
                if organization.logo
                else None
            )

            if len(courses) > 0:
                context["json_ld"] = json_ld_from(courses, organization)
            context["page_title"] = organization.name
            return context

        # If organization not found, raise 404
        raise Http404(_("Organization does not exist"))


@method_decorator(ensure_csrf_cookie, name="dispatch")
class CourseView(TemplateView):
    template_name = "public/course.html"

    def get_context_data(self, **kwargs) -> dict:  # type: ignore[no-untyped-def]
        get_token(self.request)  # Ensure CSRF token is set in cookies
        course_slug: str = kwargs.get("course_slug")  # type: ignore[assignment]
        organization_id: int = kwargs.get("organization_id")  # type: ignore[assignment]
        context = super().get_context_data(**kwargs)
        try:
            course = Course.objects.select_related("organization").get(
                slug=course_slug, organization__id=organization_id, enabled=True
            )
        except Course.DoesNotExist:
            raise Http404(_("Course does not exist"))

        course_lang_info = get_language_info(course.language)
        course_data = PublicCourseSerializer(
            id=course.id,
            title=course.title,
            slug=course.slug,
            description=course.description,
            image=self.request.build_absolute_uri(course.image.url)
            if course.image
            else None,
            imap_email=None,
            language=course.language,
            is_rtl=course_lang_info["bidi"],
            lessons=[
                content.lesson.title  # type: ignore[union-attr]
                for content in course.coursecontent_set.filter(
                    lesson__isnull=False
                ).order_by("priority")
            ],
        )
        organization_data = OrganizationSerializer(
            id=course.organization.id,
            name=course.organization.name,
            logo_url=self.request.build_absolute_uri(course.organization.logo.url)
            if course.organization.logo
            else None,
            description=course.organization.description,
            public_url=course.organization.public_url,
        )
        enroll_api_path = reverse("django_email_learning:api_public:enroll")
        current_lang_code = get_language()
        lang_info = get_language_info(current_lang_code)
        context["appContext"] = {
            "course": course_data.model_dump(),
            "organization": organization_data.model_dump(),
            "enrollApiUrl": f"{settings.DJANGO_EMAIL_LEARNING['SITE_BASE_URL']}{enroll_api_path}",
            "direction": "rtl" if lang_info["bidi"] else "ltr",
            "localeMessages": {
                "enroll_now": _("Enroll Now"),
                "enrol_for_course": _("Enroll for COURSE_NAME"),
                "email": _("email"),
                "cancel": _("Cancel"),
                "submit": _("Submit"),
                "enrollment_success": _("You are enrolled in this course."),
                "enrollment_failed": _("Enrollment failed. Please try again."),
                "email_required": _("Email is required"),
                "email_invalid": _("Please enter a valid email address"),
                "course_language": _("Course language"),
                "topics_covered": _(
                    "Here is the list of topics covered in this course:"
                ),
                "provided_by": _("Provided by ORGANIZATION_NAME"),
                "in_app_browser_or_disabled_cookies": _(
                    "It seems you are using an in-app browser or have disabled cookies. Please open this link in a regular browser and ensure cookies are enabled to enroll in courses."
                ),
                "continue": _("Continue"),
            },
        }
        context["course_title"] = course.title
        context["course_description"] = course.description
        context["course_image_url"] = (
            self.request.build_absolute_uri(course.image.url) if course.image else None
        )
        context["organization_name"] = course.organization.name
        context["organization_description"] = course.organization.description
        context["organization_logo_url"] = (
            self.request.build_absolute_uri(course.organization.logo.url)
            if course.organization.logo
            else None
        )
        context["page_title"] = course.title
        return context



from pydantic import BaseModel, Field, field_validator
import re


class EnrollmentRequest(BaseModel):
    organization_id: int
    email: str
    course_slug: str = Field(min_length=1)

    @field_validator("email")
    def validate_email(cls, v: str) -> str:
        email_regex = r"^[\w\.-]+@[\w\.-]+\.\w+$"
        if not re.match(email_regex, v):
            raise ValueError("Invalid email format")
        return v



import logging
from django.apps import apps
from django.conf.global_settings import LANGUAGES
from django.views.generic import TemplateView
from django.utils.translation import get_language_info, get_language
from django.contrib.auth.decorators import login_required
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.urls import reverse
from django_email_learning.models import Organization, OrganizationUser, Course
from django_email_learning.decorators import (
    is_platform_admin,
    is_an_organization_member,
)
from typing import Dict, Any
from django.conf import settings


DJANGO_EMAIL_LEARNING_SETTINGS: dict = getattr(settings, "DJANGO_EMAIL_LEARNING", {})
AI_CONFIGURATIONS: dict = DJANGO_EMAIL_LEARNING_SETTINGS.get("AI", {})


@method_decorator(login_required, name="dispatch")
class BasePlatformView(TemplateView):
    """Base view for all platform views with shared context"""

    def get_context_data(self, **kwargs) -> Dict[str, Any]:  # type: ignore[no-untyped-def]
        context = super().get_context_data(**kwargs)
        context.update(self.get_shared_context())
        return context

    def get_shared_context(self) -> Dict[str, Any]:
        """Get shared context for all platform views"""
        active_organization_id = self.get_or_set_active_organization()
        if self.request.user.is_superuser:
            role = "admin"
        else:
            role = OrganizationUser.objects.get(  # type: ignore[misc]
                user=self.request.user,
                organization_id=active_organization_id,
            ).role

        current_lang_code = get_language()
        lang_info = get_language_info(current_lang_code)

        return {
            "appContext": {
                "apiBaseUrl": reverse("django_email_learning:api_platform:root")[:-1],
                "platformBaseUrl": reverse("django_email_learning:platform:root")[:-1],
                "sidebarCustomComponent": {
                    "scriptUrl": DJANGO_EMAIL_LEARNING_SETTINGS.get("SIDEBAR", {})
                    .get("CUSTOM_COMPONENT", {})
                    .get("SCRIPT_URL"),
                    "componentTag": DJANGO_EMAIL_LEARNING_SETTINGS.get("SIDEBAR", {})
                    .get("CUSTOM_COMPONENT", {})
                    .get("COMPONENT_TAG"),
                    "styleUrl": DJANGO_EMAIL_LEARNING_SETTINGS.get("SIDEBAR", {})
                    .get("CUSTOM_COMPONENT", {})
                    .get("STYLE_URL"),
                },
                "userRole": role,
                "direction": "rtl" if lang_info["bidi"] else "ltr",
                "isPlatformAdmin": (
                    self.request.user.is_superuser
                    or (
                        self.request.user.is_authenticated
                        and getattr(self.request.user, "has_platform_admin_role", False)
                    )
                ),
                "aiTextEditingModel": AI_CONFIGURATIONS.get("TEXT_EDITING_MODEL"),
                "customLogo": {
                    "horizontalLight": DJANGO_EMAIL_LEARNING_SETTINGS.get("LOGO", {})
                    .get("HORIZONTAL_LOCKUP", {})
                    .get("LIGHT_BACKGROUND"),
                    "horizontalDark": DJANGO_EMAIL_LEARNING_SETTINGS.get("LOGO", {})
                    .get("HORIZONTAL_LOCKUP", {})
                    .get("DARK_BACKGROUND"),
                    "verticalLight": DJANGO_EMAIL_LEARNING_SETTINGS.get("LOGO", {})
                    .get("VERTICAL_LOCKUP", {})
                    .get("LIGHT_BACKGROUND"),
                    "verticalDark": DJANGO_EMAIL_LEARNING_SETTINGS.get("LOGO", {})
                    .get("VERTICAL_LOCKUP", {})
                    .get("DARK_BACKGROUND"),
                }
                if DJANGO_EMAIL_LEARNING_SETTINGS.get("LOGO")
                else None,
                "isOrganizationAdmin": (
                    self.request.user.is_superuser
                    or (
                        OrganizationUser.objects.filter(
                            user=self.request.user, role="admin"
                        ).exists()  # type: ignore[misc]
                    )
                ),
                "localeMessages": {
                    "organizations": _("Organizations"),
                    "course_management": _("Course Management"),
                    "learners": _("Learners"),
                    "settings": _("Settings"),
                    "api_keys": _("API Keys"),
                    "content_delivery_job": _("Content Delivery Job"),
                    "last_run": _("Last Run:"),
                    "never_run": _("This job has never been executed."),
                    "content_delivery_tooltip": _(
                        "This job should run on a regular schedule to ensure timely content delivery. Configure a cron job or cloud scheduler to execute it at appropriate intervals, such as every five minutes."
                    ),
                }
                | self.get_locale_messages(),
                "languageOptions": [
                    {"value": code, "label": name} for code, name in LANGUAGES
                ],
            },
            "activeOrganizationId": active_organization_id,
            "favicon": DJANGO_EMAIL_LEARNING_SETTINGS.get("FAVICON"),
        }

    def get_locale_messages(self) -> Dict[str, str]:
        return {}

    def get_or_set_active_organization(self) -> str:
        org = self.request.session.get("active_organization_id")
        if org:
            return org

        member = self.request.user.memberships.first()  # type: ignore[union-attr]
        logging.debug(f"User memberships: {member}")
        if member:
            org = member.organization
        elif self.request.user.is_superuser:
            org = Organization.objects.first()

        if org:
            logging.debug(f"Active organization: {org}")
            self.request.session["active_organization_id"] = str(org.id)
            return str(org.id)

        raise Exception("No active organization found for the user.")


@method_decorator(login_required, name="dispatch")
class Courses(BasePlatformView):
    template_name = "platform/courses.html"

    def get_context_data(self, **kwargs) -> dict:  # type: ignore[no-untyped-def]
        context = super().get_context_data(**kwargs)
        context["page_title"] = _("Courses")
        return context

    def get_locale_messages(self) -> Dict[str, str]:
        return {
            "actions": _("Actions"),
            "all": _("All"),
            "enable_course": _("Enable COURSE_NAME"),
            "disable_course": _("Disable COURSE_NAME"),
            "delete_course": _("Delete COURSE_NAME"),
            "enabled": _("Enabled"),
            "disabled": _("Disabled"),
            "title": _("Title"),
            "slug": _("Slug"),
            "filter": _("Filter"),
            "add_course": _("Add a Course"),
            "course_status": _("Course Status"),
            "cancel": _("Cancel"),
            "delete": _("Delete"),
            "create": _("Create"),
            "continue": _("Continue"),
            "update": _("Update"),
            "course_title": _("Course Title"),
            "course_description": _("Course Description"),
            "course_slug": _("Course Slug"),
            "course_language": _("Course Language"),
            "slug_tooltip": _(
                "The slug is a unique identifier for the course used in URLs and API endpoints. It should be lowercase, contain no spaces (use hyphens instead), and be unique across all courses for your organization. Once set, the slug cannot be changed."
            ),
            "slug_no_space": _("Slug cannot contain spaces. Use hyphens instead."),
            "add_imap_connection": _("Add IMAP Connection"),
            "imap_connection_tooltip": _(
                "You don't need an IMAP connection to build your course, but you will need one if you want your users to interact via email. For example, they can sign up or check their progress just by sending a message. This is a great solution if your audience has limited platform access."
            ),
            "new_imap_connection": _("New IMAP Connection"),
            "imap_connection": _("IMAP Connection"),
            "add": _("Add"),
            "email": _("Email"),
            "password": _("Password"),
            "server": _("Server"),
            "port": _("Port"),
            "course_enable_confirmation": _(
                "Are you sure you want to enable the course COURSE_NAME?"
            ),
            "course_disable_confirmation": _(
                "Are you sure you want to disable the course COURSE_NAME?"
            ),
            "course_delete_confirmation": _(
                "Are you sure you want to delete the course COURSE_NAME?"
            ),
            "title_required_helper_text": _("The course title is required."),
            "description_required_helper_text": _(
                "The course description is required."
            ),
            "slug_required_helper_text": _("The course slug is required."),
            "language_required_helper_text": _("The course language is required."),
            "email_required_helper_text": _("The email is required."),
            "password_required_helper_text": _("The password is required."),
            "server_required_helper_text": _("The server is required."),
            "port_required_helper_text": _("The port is required."),
            "invalid_port_helper_text": _("The port must be a valid number."),
            "invalid_email_helper_text": _("The email must be a valid email address."),
            "total_enrollments": _("Total Enrollments"),
            "upload_button_label": _("Upload Image"),
            "remove_image": _("Remove Image"),
            "uploaded_image_alt": _("Course Image"),
        }


@method_decorator(login_required, name="dispatch")
@method_decorator(is_an_organization_member(), name="dispatch")
class CourseView(BasePlatformView):
    template_name = "platform/course.html"

    def get_context_data(self, **kwargs) -> dict:  # type: ignore[no-untyped-def]
        context = super().get_context_data(**kwargs)
        course = Course.objects.get(pk=self.kwargs["course_id"])
        context["appContext"]["courseId"] = course.id
        context["appContext"]["courseTitle"] = course.title
        context["appContext"]["courseLanguage"] = course.language
        context["appContext"]["customComponent"] = None
        context["appContext"]["direction"] = (
            "rtl" if get_language_info(course.language)["bidi"] else "ltr"
        )
        context["appContext"]["showGoogleWorkspaceImport"] = bool(
            DJANGO_EMAIL_LEARNING_SETTINGS.get("GOOGLE_OAUTH_CLIENT_ID")
        ) and apps.is_installed("django_email_learning.oauth_integrations")
        context["page_title"] = _("Course: %(title)s") % {"title": course.title}
        return context

    def get_locale_messages(self) -> Dict[str, str]:
        return {
            "actions": _("Actions"),
            "enroll_learner": _("Enroll Learner"),
            "enrollment_success": _("Learner enrolled successfully."),
            "imported_from_google_success": _(
                "Learners imported from Google Workspace successfully."
            ),
            "manual_email": _("Manual Email"),
            "from_google_workspace": _("Import from Google Workspace"),
            "google_workspace_description": _(
                "If you are an administrator of a Google Workspace domain, you can import users from your domain into the platform and enroll them in this course."
            ),
            "authorize_description": _(
                "We need read-only access to your Google Workspace user directory to get started."
            ),
            "authorize_button": _("Authorize with Google"),
            "published": _("Published"),
            "type": _("Type"),
            "waiting_time": _("Waiting Time"),
            "title": _("Title"),
            "add_quiz": _("Add Quiz"),
            "send_lesson_to_yourself": _("Send it to yourself"),
            "add_lesson": _("Add Lesson"),
            "lesson": _("Lesson"),
            "quiz": _("Quiz"),
            "add": _("Add"),
            "new_lesson": _("New Lesson"),
            "update_lesson": _("Update Lesson"),
            "new_quiz": _("New Quiz"),
            "update_quiz": _("Update Quiz"),
            "editing": _("Editing..."),
            "edit_with_ai": _("Edit with AI"),
            "lesson_title": _("Lesson Title"),
            "lesson_waiting_tooltip": _(
                "Set the amount of time that we should wait after the previous lesson or quiz submission before sending this lesson"
            ),
            "upload": _("Upload"),
            "uploaded_image_preview": _("Uploaded image preview"),
            "add_image_to_editor": _("Add image to editor"),
            "remove_uploaded_image": _("Remove uploaded image"),
            "confirm_delete_uploaded_image": _("Confirm image deletion"),
            "delete_uploaded_image_warning": _(
                "Please confirm this image is not used anywhere else. Deleting it will break existing links."
            ),
            "uploaded_image_used_in_editor_error": _(
                "This image is already used in the editor content. Remove it from the content before deleting the file."
            ),
            "uploaded_image_delete_failed": _(
                "Failed to delete image file. Please try again."
            ),
            "days": _("Days"),
            "hours": _("Hours"),
            "back": _("Back"),
            "cancel": _("Cancel"),
            "delete": _("Delete"),
            "save_lesson": _("Save Lesson"),
            "lesson_saved_success": _("Lesson content saved successfully."),
            "lesson_unsaved_changes_hint": _("You have unsaved changes."),
            "save_quiz": _("Save Quiz"),
            "quiz_title": _("Quiz Title"),
            "add_question": _("Add Question"),
            "quiz_settings": _("Quiz Settings"),
            "waiting_period": _("Waiting Period"),
            "quiz_waiting_tooltip": _(
                "Time to wait after the previous content delivery before sending this quiz"
            ),
            "required_score": _("Required Score to Pass (%)"),
            "score_tooltip": _("Minimum percentage score required to pass this quiz"),
            "period": _("Period"),
            "period_tooltip": _(
                "Time to wait after the previous content delivery before sending this quiz"
            ),
            "percentage": _("Percentage"),
            "quiz_deadline": _("Deadline to Complete Quiz"),
            "deadline_tooltip": _("Maximum time allowed to complete the quiz"),
            "question_selection_strategy": _("Selection Strategy"),
            "question_selection_strategy_tooltip": _(
                "Choose how questions are selected for each quiz attempt. If the total number of questions is fewer than 6, all questions will be used even if 'Random Questions' is selected."
            ),
            "all_questions": _("All Questions"),
            "random_questions": _("Random Questions"),
            "question": _("Question"),
            "add_option": _("Add Option"),
            "option_text": _("Option Text"),
            "options": _("Options"),
            "correct_answer": _("Correct Answer"),
            "quiz_title_empty": _("Quiz title cannot be empty."),
            "at_least_one_question": _("Quiz must have at least one question"),
            "at_least_two_options": _(
                "Question QUESTION_NUMBER must have at least two answer options"
            ),
            "at_least_one_correct": _(
                "Question QUESTION_NUMBER must have at least one correct answer"
            ),
            "fix_errors": _("Please fix the errors in the form before submitting."),
            "lesson_title_required": _("Lesson title is required."),
            "lesson_content_required": _("Lesson content is required."),
            "delete_content_confirmation": _(
                "Are you sure you want to delete the content: CONTENT_TITLE?"
            ),
            "total_enrollments": _("Total Enrollments"),
            "unverified": _("Unverified"),
            "active": _("Active"),
            "deactivated": _("Deactivated"),
            "completed": _("Completed"),
            "enrollments_distribution": _("Enrollments Distribution"),
            "weekly_enrollments": _("Weekly Enrollments"),
            "unsaved_changes_warning": _(
                "You have unsaved changes. Are you sure you want to leave without saving?"
            ),
            "close_without_saving": _("Close without saving"),
        }

    def get_app_context(self) -> Dict[str, Any]:
        return {}


@method_decorator(login_required, name="dispatch")
@method_decorator(is_an_organization_member(only_admin=True), name="dispatch")
class Organizations(BasePlatformView):
    template_name = "platform/organizations.html"

    def get_context_data(self, **kwargs):  # type: ignore[no-untyped-def]
        context = super().get_context_data(**kwargs)
        context["page_title"] = _("Organizations")
        return context

    def get_locale_messages(self) -> Dict[str, str]:
        return {
            "add_organization": _("Add an Organization"),
            "actions": _("Actions"),
            "name": _("Name"),
            "description": _("Description"),
            "name_required": _("Name is required."),
            "description_required": _("Description is required."),
            "error_try_again": _("An error occurred. Please try again."),
            "upload_button_label": _("Upload Logo"),
            "create_organization": _("Create Organization"),
            "cancel": _("Cancel"),
            "delete": _("Delete"),
            "confirm_deletion": _("Confirm Deletion"),
            "remove_image": _("Remove Logo"),
            "create": _("Create"),
            "update": _("Update"),
            "uploaded_image_alt": _("Organization Logo"),
            "are_you_sure_delete_org": _(
                'Are you sure you want to delete the organization "ORGANIZATION_NAME"? All course content and users in this organization will also be deleted.'
            ),
        }


@method_decorator(login_required, name="dispatch")
@method_decorator(is_an_organization_member(only_admin=True), name="dispatch")
class SingleOrganization(BasePlatformView):
    template_name = "platform/organization.html"

    def get_context_data(self, **kwargs):  # type: ignore[no-untyped-def]
        context = super().get_context_data(**kwargs)
        organization = Organization.objects.get(pk=self.kwargs["organization_id"])
        context["organization"] = organization
        context["page_title"] = _("Organization: %(name)s") % {
            "name": organization.name
        }
        context["appContext"]["organizationId"] = organization.id
        return context

    def get_locale_messages(self) -> Dict[str, str]:
        return {
            "no_users_in_organization": _("No users in this organization yet."),
            "add_users_to_organization": _("Add user to organization"),
            "add_user": _("Add User"),
            "edit_user": _("Edit User"),
            "change_user_role": _("Change User Role"),
            "user": _("User"),
            "role": _("Role"),
            "admin": _("Admin"),
            "editor": _("Editor"),
            "viewer": _("Viewer"),
            "email": _("Email"),
            "delete_user_with_email": _("Deleting USER_EMAIL"),
            "user_delete_confirmation": _(
                "Are you sure you want to remove USER_EMAIL from this organization?"
            ),
            "actions": _("Actions"),
            "delete_note": _(
                "Note: Removing a user from this organization will not delete their account. To permanently delete the user's account, you must do so separately within the Django Admin"
            ),
            "cancel": _("Cancel"),
            "delete": _("Delete"),
        }


@method_decorator(login_required, name="dispatch")
@method_decorator(is_an_organization_member(only_admin=True), name="dispatch")
class Learners(BasePlatformView):
    template_name = "platform/learners.html"

    def get_context_data(self, **kwargs):  # type: ignore[no-untyped-def]
        context = super().get_context_data(**kwargs)
        context["page_title"] = _("Learners")
        return context

    def get_locale_messages(self) -> Dict[str, str]:
        return {
            "search_learners": _("Search learners..."),
            "learners_list": _("Learners List"),
            "nor_enrollments_found": _("No enrollments found."),
            "course": _("Course"),
            "status": _("Status"),
            "learner_registered": _("Learner Registered"),
            "learner_verified": _("Learner Verified Email"),
            "lesson_sent": _("Lesson Sent"),
            "quiz_sent": _("Quiz Sent"),
            "quiz_submitted": _("Quiz Submitted"),
            "course_completed": _("Course Completed"),
            "learner_deactivated": _("Learner Deactivated"),
            "score": _("Score"),
            "result": _("Result"),
            "reason": _("Reason"),
            "passed": _("Passed"),
            "failed": _("Failed"),
            "enrollment_details": _("Enrollment Details"),
            "enrollment_id": _("Enrollment ID"),
            "unverified": _("Unverified"),
            "active": _("Active"),
            "completed": _("Completed"),
            "deactivated": _("Deactivated"),
            "canceled": _("Canceled"),
            "blcoked": _("Blocked"),
            "inactive": _("Inactive"),
        }


@method_decorator(login_required, name="dispatch")
@method_decorator(is_platform_admin(), name="dispatch")
class ApiKeys(BasePlatformView):
    template_name = "platform/settings_api_keys.html"

    def get_context_data(self, **kwargs):  # type: ignore[no-untyped-def]
        context = super().get_context_data(**kwargs)
        context["page_title"] = _("API Keys")
        return context

    def get_locale_messages(self) -> Dict[str, str]:
        return {
            "settings": _("Settings"),
            "api_keys": _("API Keys"),
            "add_api_key": _("Add API Key"),
            "display_key": _("Display Key"),
            "hide_key": _("Hide Key"),
            "actions": _("Actions"),
            "key": _("Key"),
            "created_at": _("Created At"),
            "delete": _("Delete"),
            "are_you_sure_delete_key": _(
                "Are you sure you want to delete this API key?"
            ),
            "created_by": _("Created By"),
            "cancel": _("Cancel"),
            "confirm_deletion": _("Confirm Deletion"),
            "api_key_intro": _(
                "API keys allow external applications to interact with the platform and execute jobs. This is ideal for using cloud scheduling or third-party integrations instead of managing local cron jobs. You can create, view, and manage your keys below."
            ),
        }



from django.urls import path

from .views import RedirectView, SessionView, SessionsView

app_name = "oauth_integrations"

urlpatterns = [
    path("sessions/", SessionsView.as_view(), name="sessions_view"),
    path("sessions/<str:session_id>/", SessionView.as_view(), name="session_view"),
    path("redirect/", RedirectView.as_view(), name="redirect_view"),
]



import json

from django.http import JsonResponse, HttpResponse
from django.shortcuts import render
from django.utils.translation import get_language, get_language_info
from django.utils.translation import gettext as _
from django.views import View
from pydantic import ValidationError

from django_email_learning.apps import PLATFORM_ADMIN_GROUP_NAME
from django_email_learning.models import Course, OrganizationUser
from django_email_learning.services.command_models.enroll_from_google_directory_command import (
    EnrollFromGoogleDirectoryCommand,
)
from django_email_learning.services.command_handler_service import CommandHandlerService
from django_email_learning.services.jwt_service import (
    InvalidTokenException,
    decode_jwt,
    generate_jwt,
)

from .models import Session, SessionState
from .serializers import CreateSessionRequest


def _command_result_response(  # type: ignore[no-untyped-def]
    request,
    *,
    page_title: str,
    success_message: str | None = None,
    error_message: str | None = None,
    status_code: int = 200,
) -> HttpResponse:
    current_lang_code = get_language()
    lang_info = get_language_info(current_lang_code)
    return render(
        request,
        "personalised/command_result.html",
        {
            "page_title": page_title,
            "appContext": {
                "successMessage": success_message,
                "errorMessage": error_message,
                "closeWindow": True,
                "direction": "rtl" if lang_info["bidi"] else "ltr",
                "localeMessages": {"Confirm": _("Confirm")},
            },
        },
        status=status_code,
    )


def _has_oauth_session_access(user, organization_id: int) -> bool:  # type: ignore[no-untyped-def]
    if user.is_superuser:
        return True
    if user.groups.filter(name=PLATFORM_ADMIN_GROUP_NAME).exists():
        return True
    return OrganizationUser.objects.filter(
        user=user,
        organization_id=organization_id,
        role="admin",
    ).exists()


class SessionsView(View):
    def post(self, request, *args, **kwargs):  # type: ignore[no-untyped-def]
        if not request.user.is_authenticated:
            return JsonResponse({"error": "Unauthorized"}, status=401)

        payload = json.loads(request.body)

        try:
            serializer = CreateSessionRequest.model_validate(payload)
        except ValidationError as ve:
            return JsonResponse({"error": ve.errors()}, status=400)

        command = serializer.request.command
        if not isinstance(command, EnrollFromGoogleDirectoryCommand):
            return JsonResponse(
                {"error": "Unsupported command for oauth session"}, status=400
            )

        try:
            course = Course.objects.get(id=command.course_id)
        except Course.DoesNotExist:
            return JsonResponse({"error": "Course not found"}, status=404)

        if not _has_oauth_session_access(request.user, course.organization_id):
            return JsonResponse({"error": "Forbidden"}, status=403)

        temp_session = Session.objects.create(jwt_token="pending")
        authorization_url = command.get_authorization_url(temp_session.session_id)
        command_payload = serializer.request.model_dump(mode="json")
        temp_session.jwt_token = generate_jwt(command_payload)
        temp_session.save(update_fields=["jwt_token"])

        return JsonResponse(
            {
                "session_id": temp_session.session_id,
                "authorization_url": authorization_url,
            },
            status=201,
        )


class SessionView(View):
    def get(self, request, *args, **kwargs):  # type: ignore[no-untyped-def]
        try:
            session = Session.objects.get(session_id=kwargs["session_id"])
            return JsonResponse(
                {"session_id": session.session_id, "state": session.state}, status=200
            )
        except Session.DoesNotExist:
            return JsonResponse({"error": "Session not found"}, status=404)


class RedirectView(View):
    def get(self, request, *args, **kwargs):  # type: ignore[no-untyped-def]
        session_id = request.GET.get("state")
        code = request.GET.get("code")
        if not session_id:
            return _command_result_response(
                request,
                page_title=_("Authorization Error"),
                error_message=_("Missing state parameter."),
                status_code=400,
            )

        try:
            session = Session.objects.get(session_id=session_id)
        except Session.DoesNotExist:
            return _command_result_response(
                request,
                page_title=_("Authorization Error"),
                error_message=_("Invalid session identifier."),
                status_code=404,
            )

        if not code:
            session.state = SessionState.FAILED
            session.save(update_fields=["state"])
            return _command_result_response(
                request,
                page_title=_("Authorization Error"),
                error_message=_("Missing code parameter."),
                status_code=400,
            )

        try:
            session.state = SessionState.PROCESSING
            session.save(update_fields=["state"])

            decoded_request = decode_jwt(session.jwt_token)
            command_payload = decoded_request.get("command", {})
            command_payload["code"] = code
            command_payload["state"] = session_id
            decoded_request["command"] = command_payload
            command_handler = CommandHandlerService()
            command_handler.handle_json_command(decoded_request)
            session.state = SessionState.COMPLETED
            session.save(update_fields=["state"])
            return _command_result_response(
                request,
                page_title=_("Authorization Complete"),
                success_message=_(
                    "Google authorization completed successfully. You can close this window."
                ),
                status_code=200,
            )
        except InvalidTokenException:
            session.state = SessionState.FAILED
            session.save(update_fields=["state"])
            return _command_result_response(
                request,
                page_title=_("Authorization Error"),
                error_message=_("Invalid OAuth session token."),
                status_code=400,
            )
        except Exception as e:  # noqa: BLE001
            session.state = SessionState.FAILED
            session.save(update_fields=["state"])
            return _command_result_response(
                request,
                page_title=_("Authorization Error"),
                error_message=str(e),
                status_code=400,
            )



from django.conf import settings
from django.utils.module_loading import import_string


class MetricsService:
    def __init__(self) -> None:
        DJANGO_EMAIL_LEARNING_SETTINGS: dict = getattr(
            settings, "DJANGO_EMAIL_LEARNING", {}
        )
        try:
            configured_metric_recorder = import_string(
                DJANGO_EMAIL_LEARNING_SETTINGS["METRIC_RECORDER"]
            )
            self.metric_recorder = (
                configured_metric_recorder()
                if isinstance(configured_metric_recorder, type)
                else configured_metric_recorder
            )
        except KeyError:
            from django_email_learning.services.defaults.log_based_metric_recorder import (
                LogBasedMetricRecorder,
            )

            self.metric_recorder = LogBasedMetricRecorder()

    def user_enrolled_in_course(self, course_slug: str, organization_id: int) -> None:
        self.metric_recorder.user_enrolled_in_course(course_slug, organization_id)

    def quiz_sent(self, course_slug: str, organization_id: int, quiz_id: int) -> None:
        self.metric_recorder.quiz_sent(course_slug, organization_id, quiz_id)

    def lesson_sent(
        self, course_slug: str, organization_id: int, lesson_id: int
    ) -> None:
        self.metric_recorder.lesson_sent(course_slug, organization_id, lesson_id)

    def user_enrollment_activated(self, course_slug: str, organization_id: int) -> None:
        self.metric_recorder.user_enrollment_activated(course_slug, organization_id)

    def user_enrollment_deactivated(
        self, course_slug: str, organization_id: int, reason: str
    ) -> None:
        self.metric_recorder.user_enrollment_deactivated(
            course_slug, organization_id, reason
        )

    def user_completed_course(self, course_slug: str, organization_id: int) -> None:
        self.metric_recorder.user_completed_course(course_slug, organization_id)

    def quiz_submitted(
        self, course_slug: str, organization_id: int, quiz_id: int, is_passed: bool
    ) -> None:
        self.metric_recorder.quiz_submitted(
            course_slug, organization_id, quiz_id, is_passed
        )

    def method_executed(self, method_name: str, execution_time: int) -> None:
        self.metric_recorder.method_executed(method_name, execution_time)

    def delivery_schedule_blocked(self, content_id: int) -> None:
        self.metric_recorder.delivery_schedule_blocked(content_id)



import logging
from django_email_learning.ports.email_sender_protocol import EmailSenderProtocol
from django_email_learning.services.utils import mask_email
from django.core.mail import EmailMultiAlternatives

logger = logging.getLogger(__name__)


class DjangoEmailSender(EmailSenderProtocol):
    def _mask_recipients(self, recipients: list[str]) -> str:
        """Mask all recipient email addresses for logging."""
        if not recipients:
            return "no recipients"
        masked = [mask_email(recipient) for recipient in recipients]
        return ", ".join(masked)

    def send_email(self, email: EmailMultiAlternatives) -> None:
        masked_recipients = self._mask_recipients(email.to)
        try:
            logger.info(f"Sending email to {masked_recipients}")
            email.send()
            logger.info(f"Email sent successfully to {masked_recipients}")
        except Exception as e:
            logger.error(f"Failed to send email to {masked_recipients}: {str(e)}")
            raise



from django.urls import path
from django.views.defaults import page_not_found
from django_email_learning.platform.api.views import (
    ApiKeyView,
    GetOrCreateUserByEmail,
    SendLessonToPlatformUser,
    SingleApiKeyView,
    CourseView,
    EnrollmentsView,
    EnrollmentView,
    EnrollmentsStatisticsView,
    FileView,
    ImapConnectionView,
    OrganizationsView,
    OrganizationUsersView,
    SingleOrganizationUserView,
    SingleOrganizationView,
    SingleCourseView,
    CourseContentView,
    ReorderCourseContentView,
    SingleCourseContentView,
    UpdateSessionView,
    LearnersView,
    SingleLearnerView,
    JobsStatus,
)

app_name = "django_email_learning"

urlpatterns = [
    path(
        "organizations/<int:organization_id>/courses/",
        CourseView.as_view(),
        name="course_view",
    ),
    path(
        "organizations/<int:organization_id>/users/",
        OrganizationUsersView.as_view(),
        name="organization_users_view",
    ),
    path(
        "organizations/<int:organization_id>/users/<int:user_id>/",
        SingleOrganizationUserView.as_view(),
        name="single_organization_user_view",
    ),
    path(
        "organizations/<int:organization_id>/imap-connections/",
        ImapConnectionView.as_view(),
        name="imap_connection_view",
    ),
    path(
        "organizations/<int:organization_id>/courses/<int:course_id>/",
        SingleCourseView.as_view(),
        name="single_course_view",
    ),
    path(
        "organizations/<int:organization_id>/courses/<int:course_id>/contents/",
        CourseContentView.as_view(),
        name="course_content_view",
    ),
    path(
        "organizations/<int:organization_id>/courses/<int:course_id>/contents/reorder/",
        ReorderCourseContentView.as_view(),
        name="reorder_course_contents_view",
    ),
    path(
        "organizations/<int:organization_id>/courses/<int:course_id>/contents/<int:course_content_id>/",
        SingleCourseContentView.as_view(),
        name="single_course_content_view",
    ),
    path(
        "organizations/<int:organization_id>/learners/",
        LearnersView.as_view(),
        name="learners_view",
    ),
    path(
        "organizations/<int:organization_id>/learners/<int:learner_id>/",
        SingleLearnerView.as_view(),
        name="single_learner_view",
    ),
    path(
        "organizations/<int:organization_id>/enrollments/<int:enrollment_id>/",
        EnrollmentView.as_view(),
        name="enrollment_view",
    ),
    path(
        "organizations/<int:organization_id>/courses/<int:course_id>/enrollments/",
        EnrollmentsView.as_view(),
        name="enrollments_view",
    ),
    path(
        "organizations/<int:organization_id>/courses/<int:course_id>/enrollments/statistics/",
        EnrollmentsStatisticsView.as_view(),
        name="enrollments_statistics_view",
    ),
    path(
        "organizations/<int:organization_id>/send-lesson/",
        SendLessonToPlatformUser.as_view(),
        name="send_lesson_to_platform_user_view",
    ),
    path(
        "organizations/<int:organization_id>/file/",
        FileView.as_view(),
        name="file_view",
    ),
    path("organizations/", OrganizationsView.as_view(), name="organizations_view"),
    path(
        "organizations/<int:organization_id>/",
        SingleOrganizationView.as_view(),
        name="single_organization_view",
    ),
    path("status/jobs/", JobsStatus.as_view(), name="jobs_status_view"),
    path("api_keys/", ApiKeyView.as_view(), name="api_key_view"),
    path(
        "api_keys/<int:api_key_id>/",
        SingleApiKeyView.as_view(),
        name="single_api_key_view",
    ),
    path(
        "users/get-or-create-by-email/",
        GetOrCreateUserByEmail.as_view(),
        name="get_or_create_user_by_email_view",
    ),
    path("session", UpdateSessionView.as_view(), name="update_session_view"),
    path("", page_not_found, name="root"),
]



from django_email_learning.services.command_models.abstract_command import (
    AbstractCommand,
)
from django_email_learning.models import Enrollment, EnrollmentStatus
from pydantic import Field

from django_email_learning.services.command_models.exceptions.invalid_enrollment_error import (
    InvalidEnrollmentError,
)
from django_email_learning.services.command_models.exceptions.invalid_verification_code_error import (
    InvalidVerificationCodeError,
)
from django_email_learning.services.email_sender_service import EmailSenderService
from django_email_learning.services.metrics_service import MetricsService
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string
from django.utils.translation import gettext as _
from django.conf import settings
from typing import Literal


class VerifyEnrollmentCommand(AbstractCommand):
    command_name: Literal["verify_enrollment"] = "verify_enrollment"
    enrollment_id: int = Field(..., gt=0)
    verification_code: str = Field(..., pattern=r"^\d{6}$")

    def execute(self) -> None:
        metric_service = MetricsService()
        try:
            enrollment = Enrollment.objects.get(
                id=self.enrollment_id, status=EnrollmentStatus.UNVERIFIED
            )
        except Enrollment.DoesNotExist:
            self.logger.error(
                f"Verification Failed: No unverified enrollment found with ID {self.enrollment_id}"
            )
            # Check if enrollment exists but is not unverified
            if Enrollment.objects.filter(id=self.enrollment_id).exists():
                return
            raise InvalidEnrollmentError(
                f"No unverified enrollment found with ID {self.enrollment_id}"
            )

        if str(enrollment.activation_code) != str(self.verification_code):
            self.logger.error(
                f"Verification Failed: Invalid verification code for Enrollment ID {self.enrollment_id}"
            )
            raise InvalidVerificationCodeError(
                f"Invalid verification code for Enrollment ID {self.enrollment_id}"
            )

        enrollment.status = EnrollmentStatus.ACTIVE
        enrollment.activation_code = None
        enrollment.save()
        self.logger.info(
            f"Enrollment Verified: Enrollment ID {self.enrollment_id} has been activated"
        )

        enrollment.schedule_first_content_delivery()
        self.logger.info(
            f"Content Delivery Scheduled: First content delivery scheduled for Enrollment ID {self.enrollment_id}"
        )
        metric_service.user_enrollment_activated(
            enrollment.course.slug, enrollment.course.organization.id
        )

        # Send confirmation email
        email_service = EmailSenderService()
        subject = _("Enrollment Verified")
        course_image_url = None
        if enrollment.course.image:
            image_url = enrollment.course.image.url
            if image_url.startswith(("http://", "https://")):
                course_image_url = image_url
            else:
                site_base_url = settings.DJANGO_EMAIL_LEARNING["SITE_BASE_URL"]
                course_image_url = f"{site_base_url}".rstrip("/") + image_url

        body = render_to_string(
            "emails/enrollment_verified.txt",
            {
                "course_title": enrollment.course.title,
                "organization_name": enrollment.course.organization.name,
            },
        )

        email = EmailMultiAlternatives(
            subject=subject,
            body=body,
            from_email=email_service.from_email,
            to=[enrollment.learner.email],
        )
        html_content = render_to_string(
            "emails/enrollment_verified.html",
            {
                "course_title": enrollment.course.title,
                "organization_name": enrollment.course.organization.name,
                "course_image_url": course_image_url,
            },
        )
        email.attach_alternative(html_content, "text/html")

        email_service.send(email=email)



from .ai_service_protocol import AiServiceProtocol
from openai import OpenAI
from django.conf import settings
import os


DJANGO_EMAIL_LEARNING_CONFIG = getattr(settings, "DJANGO_EMAIL_LEARNING", {})


class OpenAiAdapter(AiServiceProtocol):
    def __init__(self) -> None:
        API_KEY = DJANGO_EMAIL_LEARNING_CONFIG.get("AI", {}).get(
            "OPENAI_API_KEY", os.getenv("OPENAI_API_KEY")
        )
        if not API_KEY:
            raise ValueError(
                "OpenAI API key is required. Please set it in Django settings or as an environment variable."
            )
        self._client = OpenAI(api_key=API_KEY)

    def edit_text(self, text: str, model: str) -> str:
        # Placeholder for OpenAI API integration
        response = self._client.responses.create(
            model=model,
            input=[
                {
                    "role": "developer",
                    "content": "You are a helpful assistant that edits text for grammar and clarity. Preserve the existing HTML structure and tags when present. Do not add wrapper containers like <ul>, <ol>, or <p> unless they already exist in the input. Return ONLY the edited text.",
                },
                {"role": "user", "content": text},
            ],
            store=False,
        )
        return response.output_text



from django.urls import path
from django_email_learning.personalised.views import (
    QuizPublicView,
    VerifyEnrollmentView,
    CertificateFormView,
    CertificateView,
    UnsubscribeView,
)

app_name = "django_email_learning"

urlpatterns = [
    path("quiz/", QuizPublicView.as_view(), name="quiz_public_view"),
    path(
        "verify-enrollment/", VerifyEnrollmentView.as_view(), name="verify_enrollment"
    ),
    path("certificate-form/", CertificateFormView.as_view(), name="certificate_form"),
    path(
        "certificate/<str:certificate_number>/",
        CertificateView.as_view(),
        name="certificate",
    ),
    path("unsubscribe/", UnsubscribeView.as_view(), name="unsubscribe"),
]



from django.urls import path
from django_email_learning.public.views import OrganizationView, CourseView

app_name = "django_email_learning"

urlpatterns = [
    path(
        "organizations/<int:organization_id>/",
        OrganizationView.as_view(),
        name="organization_view",
    ),
    path(
        "organizations/<int:organization_id>/courses/<slug:course_slug>/",
        CourseView.as_view(),
        name="course_view",
    ),
]



import re
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    field_serializer,
    field_validator,
    model_validator,
)
from datetime import datetime
from typing import Optional, Literal, Any, Callable
from django.urls import reverse
from django_email_learning.models import (
    ApiKey,
    DeliveryStatus,
    Organization,
    ImapConnection,
    Lesson,
    Quiz,
    Question,
    Answer,
    CourseContent,
    Course,
    QuizSelectionStrategy,
    Enrollment,
    EnrollmentStatus,
    OrganizationUser,
)
from django_email_learning.services.jwt_service import generate_jwt
from django.utils.translation import get_language_info
import enum


class ApiKeyResponse(BaseModel):
    id: int
    key: str
    created_at: datetime
    created_by: Optional[str] = None

    @staticmethod
    def from_django_model(api_key: ApiKey) -> "ApiKeyResponse":
        decrypted_key = api_key.decrypt_password(api_key.key)
        salt = api_key.salt
        jwt_key = generate_jwt({"key": decrypted_key, "salt": salt}, exp=datetime.max)

        return ApiKeyResponse.model_validate(
            {
                "id": api_key.id,  # type: ignore[attr-defined]
                "key": jwt_key,
                "created_at": api_key.created_at,
                "created_by": api_key.created_by.username
                if api_key.created_by
                else None,
            }
        )


class GetOrCreateUserRequest(BaseModel):
    email: str = Field(min_length=1, examples=["user@example.com"])
    organization_id: int = Field(gt=0, examples=[1])

    @field_validator("email")
    def validate_email(cls, email: str) -> str:
        email_regex = r"^[\w\.-]+@[\w\.-]+\.\w+$"
        if not re.match(email_regex, email):
            raise ValueError("Invalid email format")
        return email


class CreateEnrollmentRequest(BaseModel):
    learner_email: str = Field(min_length=1, examples=["user@example.com"])


class UserResponse(BaseModel):
    id: int
    email: str

    model_config = ConfigDict(from_attributes=True)


class Identifier(BaseModel):
    id: int


class CreateCourseRequest(BaseModel):
    title: str = Field(min_length=1, examples=["Introduction to Python"])
    slug: str = Field(
        min_length=1,
        examples=["intro-to-python"],
        description="A short label for the course, used in URLs or email interactive actions. "
        "You can not edit it later.",
    )
    description: Optional[str] = Field(
        None, examples=["A beginner's course on Python programming."]
    )
    imap_connection_id: Optional[int] = Field(None, examples=[1])
    image: Optional[str] = Field(None, examples=["/path/to/course_image.png"])
    language: str = Field(min_length=2, max_length=10, examples=["en"])

    def to_django_model(self, organization_id: int) -> Course:
        organization = Organization.objects.get(id=organization_id)
        if not organization:
            raise ValueError(f"Organization with id {organization_id} does not exist.")
        imap_connection = None
        if self.imap_connection_id:
            try:
                imap_connection = ImapConnection.objects.get(
                    id=self.imap_connection_id, organization=organization
                )
            except ImapConnection.DoesNotExist:
                raise ValueError(
                    f"ImapConnection with id {self.imap_connection_id} does not exist."
                )
            imap_connection = ImapConnection.objects.get(
                id=self.imap_connection_id, organization=organization
            )
        course = Course(
            title=self.title,
            slug=self.slug,
            description=self.description,
            organization=organization,
            language=self.language,
        )
        if imap_connection:
            course.imap_connection = imap_connection
        if self.image:
            course.replace_image(self.image)
        return course


class UpdateCourseRequest(BaseModel):
    model_config = ConfigDict(extra="forbid")
    title: Optional[str] = Field(
        None, min_length=1, examples=["Introduction to Python"]
    )
    description: Optional[str] = Field(
        None, examples=["A beginner's course on Python programming."]
    )
    imap_connection_id: Optional[int] = Field(None, examples=[1])
    enabled: Optional[bool] = Field(None, examples=[True])
    reset_imap_connection: Optional[bool] = Field(None, examples=[False])
    image: Optional[str] = Field(None, examples=["/path/to/course_image.png"])
    language: Optional[str] = Field(None, min_length=2, max_length=10, examples=["en"])

    def to_django_model(self, course_id: int) -> Course:
        try:
            course = Course.objects.get(id=course_id)
        except Course.DoesNotExist:
            raise ValueError(f"Course with id {course_id} does not exist.")
        if self.reset_imap_connection and self.imap_connection_id is not None:
            raise ValueError(
                "Cannot set imap_connection_id when reset_imap_connection is True."
            )

        if self.title is not None:
            course.title = self.title
        if self.description is not None:
            course.description = self.description
        if self.imap_connection_id is not None:
            imap_connection = ImapConnection.objects.get(id=self.imap_connection_id)
            course.imap_connection = imap_connection
        if self.enabled is not None:
            course.enabled = self.enabled
        if self.reset_imap_connection:
            course.imap_connection = None
        if self.image is not None:
            if self.image != "SKIP":
                course.replace_image(self.image)
        if not self.image:
            course.image = None
        if self.language is not None:
            course.language = self.language
        return course


class CourseResponse(BaseModel):
    id: int
    title: str
    slug: str
    description: Optional[str]
    organization_id: int
    imap_connection_id: Optional[int]
    enabled: bool
    enrollments_count: dict[str, int]
    image: Optional[str] = None
    image_path: Optional[str] = None
    language: str
    is_rtl: bool = False

    model_config = ConfigDict(from_attributes=True)

    @staticmethod
    def from_django_model(
        course: Course, abs_url_builder: Callable
    ) -> "CourseResponse":
        language_info = get_language_info(course.language)
        return CourseResponse.model_validate(
            {
                "id": course.id,
                "title": course.title,
                "slug": course.slug,
                "description": course.description,
                "organization_id": course.organization.id,
                "imap_connection_id": course.imap_connection.id
                if course.imap_connection
                else None,
                "enabled": course.enabled,
                "enrollments_count": course.enrollments_count,
                "image": abs_url_builder(course.image.url) if course.image else None,
                "image_path": course.image.name if course.image else None,
                "language": course.language,
                "is_rtl": language_info["bidi"],
            }
        )


class CourseSummaryResponse(BaseModel):
    id: int
    title: str
    slug: str

    model_config = ConfigDict(from_attributes=True)


class CreateImapConnectionRequest(BaseModel):
    email: str = Field(min_length=1, examples=["user@example.com"])
    password: str = Field(min_length=1, examples=["aSafePassword123!"])
    server: str = Field(min_length=1, examples=["imap.example.com"])
    port: int = Field(gt=0, examples=[993])

    def to_django_model(self, organization_id: int) -> ImapConnection:
        organization = Organization.objects.get(id=organization_id)
        if not organization:
            raise ValueError(f"Organization with id {organization_id} does not exist.")
        imap_connection = ImapConnection(
            email=self.email,
            password=self.password,
            server=self.server,
            port=self.port,
            organization=organization,
        )
        return imap_connection


class ImapConnectionResponse(BaseModel):
    id: int
    email: str
    server: str
    port: int
    organization_id: int

    model_config = ConfigDict(from_attributes=True)


class OrganizationResponse(BaseModel):
    id: int
    name: str
    logo: Optional[str] = None
    logo_path: Optional[str] = None
    description: Optional[str] = None
    public_url: str

    model_config = ConfigDict(from_attributes=True)

    @staticmethod
    def from_django_model(
        organization: Organization, abs_url_builder: Callable
    ) -> "OrganizationResponse":
        url = reverse(
            "django_email_learning:public:organization_view",
            kwargs={"organization_id": organization.id},
        )
        return OrganizationResponse.model_validate(
            {
                "id": organization.id,
                "name": organization.name,
                "logo": abs_url_builder(organization.logo.url)
                if organization.logo
                else None,
                "logo_path": organization.logo.name if organization.logo else None,
                "description": organization.description,
                "public_url": abs_url_builder(url),
            }
        )


class CreateOrganizationRequest(BaseModel):
    name: str = Field(min_length=1, examples=["AvaCode"])
    description: Optional[str] = Field(
        None, examples=["A description of the organization."]
    )
    logo: Optional[str] = Field(None, examples=["/path/to/logo.png"])

    def to_django_model(self) -> Organization:
        organization = Organization(name=self.name, description=self.description)
        organization.save()
        organization.refresh_from_db()
        if self.logo:
            organization.replace_logo(self.logo)

        return organization


class UpdateOrganizationRequest(BaseModel):
    model_config = ConfigDict(extra="forbid")
    name: Optional[str] = Field(None, min_length=1, examples=["AvaCode"])
    description: Optional[str] = Field(
        None, examples=["A description of the organization."]
    )
    logo: Optional[str] = Field(None, examples=["/path/to/logo.png"])
    remove_logo: Optional[bool] = Field(None, examples=[True])


class UserRole(enum.StrEnum):
    ADMIN = "admin"
    EDITOR = "editor"
    VIEWER = "viewer"


class AddOrganizationUserRequest(BaseModel):
    user_id: int = Field(gt=0, examples=[1])
    role: UserRole = Field(min_length=1, examples=[UserRole.ADMIN])


class UpdateOrganizationUserRoleRequest(BaseModel):
    role: UserRole = Field(min_length=1, examples=[UserRole.ADMIN])


class OrganizationUserResponse(BaseModel):
    id: int
    user_id: int
    organization_id: int
    email: str
    role: UserRole

    @staticmethod
    def from_django_model(org_user: OrganizationUser) -> "OrganizationUserResponse":
        return OrganizationUserResponse(
            id=org_user.id,
            user_id=org_user.user.id,
            organization_id=org_user.organization.id,
            email=org_user.user.email,
            role=UserRole(org_user.role),
        )


class UpdateSessionRequest(BaseModel):
    active_organization_id: int = Field(examples=[1])

    model_config = ConfigDict(extra="forbid")


class SessionInfo(BaseModel):
    active_organization_id: int

    @classmethod
    def populate_from_session(cls, session):  # type: ignore[no-untyped-def]
        return super().model_validate(
            {"active_organization_id": session.get("active_organization_id")}
        )


class LessonCreate(BaseModel):
    title: str
    content: str
    type: Literal["lesson"] = "lesson"


class LessonUpdate(BaseModel):
    title: Optional[str] = None
    content: Optional[str] = None

    model_config = ConfigDict(extra="forbid")


class LessonResponse(BaseModel):
    id: int
    title: str
    content: str

    model_config = ConfigDict(from_attributes=True)


class AnswerCreate(BaseModel):
    text: str
    is_correct: bool = Field(examples=[True])


class AnswerUpdate(AnswerCreate):
    id: Optional[int] = None


class AnswerObject(BaseModel):
    id: int
    text: str
    is_correct: bool

    model_config = ConfigDict(from_attributes=True)


class QuestionCreate(BaseModel):
    text: str
    priority: int = Field(gt=0, examples=[1])
    answers: list[AnswerCreate] = Field(min_length=2)

    @field_validator("answers")
    @classmethod
    def at_least_one_correct_answer(
        cls, answers: list[AnswerCreate]
    ) -> list[AnswerCreate]:
        correct_answers = [answer for answer in answers if answer.is_correct]
        if not correct_answers:
            raise ValueError("At least one answer must be marked as correct.")
        return answers


class QuestionUpdate(QuestionCreate):
    id: Optional[int] = None
    answers: list[AnswerUpdate] = Field(min_length=2)  # type: ignore[assignment]


class QuestionObject(BaseModel):
    id: int
    text: str
    priority: int
    answers: Any  # Will be converted to list in field_serializer

    @field_serializer("answers")
    def serialize_answers(self, answers: Any) -> list[dict]:
        return [
            AnswerObject.model_validate(answer).model_dump() for answer in answers.all()
        ]

    model_config = ConfigDict(from_attributes=True)


MIN_QUIZ_DEADLINE = 1
MAX_QUIZ_DEADLINE = 30


class UpdateQuiz(BaseModel):
    questions: Optional[list[QuestionUpdate]] = Field(min_length=1)
    title: Optional[str] = None
    required_score: Optional[int] = Field(ge=0, examples=[80], default=None)
    selection_strategy: QuizSelectionStrategy
    deadline_days: int = Field(
        ge=MIN_QUIZ_DEADLINE, le=MAX_QUIZ_DEADLINE, examples=[14]
    )

    model_config = ConfigDict(extra="forbid")


class QuizCreate(BaseModel):
    title: str
    required_score: int = Field(ge=0, examples=[80])
    selection_strategy: QuizSelectionStrategy
    deadline_days: int = Field(
        ge=MIN_QUIZ_DEADLINE, le=MAX_QUIZ_DEADLINE, examples=[14]
    )
    questions: list[QuestionCreate] = Field(min_length=1)
    type: Literal["quiz"] = "quiz"


class QuizResponse(BaseModel):
    id: int
    title: str
    required_score: int
    selection_strategy: str
    deadline_days: int = Field(ge=MIN_QUIZ_DEADLINE, le=MAX_QUIZ_DEADLINE)
    questions: Any  # Will be converted to list in field_serializer

    @field_serializer("questions")
    def serialize_questions(self, questions: Any) -> list[dict]:
        return [
            QuestionObject.model_validate(question).model_dump()
            for question in questions.all()
        ]

    model_config = ConfigDict(from_attributes=True)


class PeriodType(enum.StrEnum):
    HOURS = "hours"
    DAYS = "days"


class WaitingPeriod(BaseModel):
    period: int = Field(gt=0, examples=[7])
    type: PeriodType

    def to_seconds(self) -> int:
        if self.type == PeriodType.HOURS:
            return self.period * 3600
        elif self.type == PeriodType.DAYS:
            return self.period * 86400
        else:
            raise ValueError(f"Unsupported period type: {self.type}")

    @classmethod
    def from_seconds(cls, seconds: int) -> "WaitingPeriod":
        if seconds % 86400 == 0:
            return cls(period=seconds // 86400, type=PeriodType.DAYS)
        elif seconds % 3600 == 0:
            return cls(period=seconds // 3600, type=PeriodType.HOURS)
        else:
            raise ValueError(
                f"Cannot convert {seconds} seconds to a valid WaitingPeriod."
            )


class EnrollmentSummaryResponse(BaseModel):
    model_config = ConfigDict(from_attributes=True)
    id: int
    course_title: str
    status: EnrollmentStatus
    progress: int
    certificate_url: str | None = None


class EnrollmentsCount(BaseModel):
    total: int
    completed: int


class LearnerResponse(BaseModel):
    model_config = ConfigDict(from_attributes=True)
    id: int
    email: str
    enrollments_count: EnrollmentsCount


class EventType(enum.StrEnum):
    REGISTERED = "registered"
    VERIFIED = "verified"
    DEACTIVATED = "deactivated"
    QUIZ_SUBMITED = "quiz_submitted"
    CONTENT_SENT = "content_sent"
    COURSE_COMPLETED = "course_completed"


class DeactivatedEvent(BaseModel):
    type: Literal[EventType.DEACTIVATED] = Field(
        default=EventType.DEACTIVATED, exclude=True
    )
    reason: str


class QuizSubmitedEvent(BaseModel):
    type: Literal[EventType.QUIZ_SUBMITED] = Field(
        default=EventType.QUIZ_SUBMITED, exclude=True
    )
    quiz_id: int
    quiz_title: str
    score: int
    is_passed: bool
    attempt_number: int


class ContentSentEvent(BaseModel):
    type: Literal[EventType.CONTENT_SENT] = Field(
        default=EventType.CONTENT_SENT, exclude=True
    )
    course_content_id: int
    course_content_title: str
    course_content_type: str


class Event(BaseModel):
    type: EventType
    timestamp: datetime
    event_data: DeactivatedEvent | QuizSubmitedEvent | ContentSentEvent | None = Field(
        discriminator="type"
    )  # REGISTERED, VERIFIED, COURSE_COMPLETED have no additional data


class EnrollmentResponse(BaseModel):
    id: int
    learner: LearnerResponse
    course: CourseSummaryResponse
    status: EnrollmentStatus
    events: list[Event]

    @staticmethod
    def from_django_model(enrollment: Enrollment) -> "EnrollmentResponse":
        events = [
            Event(
                type=EventType.REGISTERED,
                timestamp=enrollment.enrolled_at,
                event_data=None,
            )
        ]
        if enrollment.activated_at:
            events.append(
                Event(
                    type=EventType.VERIFIED,
                    timestamp=enrollment.activated_at,
                    event_data=None,
                )
            )
        for delivery in enrollment.content_deliveries.all():  # type: ignore[attr-defined]
            schedule_no = 0
            for schedule in delivery.delivery_schedules.filter(
                status=DeliveryStatus.DELIVERED
            ):
                schedule_no += 1
                events.append(
                    Event(
                        type=EventType.CONTENT_SENT,
                        timestamp=schedule.delivered_at,  # type: ignore[arg-type]
                        event_data=ContentSentEvent(
                            course_content_id=delivery.course_content.id,
                            course_content_title=delivery.course_content.lesson.title
                            if delivery.course_content.lesson
                            else delivery.course_content.quiz.title,  # type: ignore[union-attr]
                            course_content_type=delivery.course_content.type,
                        ),
                    )
                )
                if delivery.course_content.type == "quiz":
                    attempt_number = 0
                    quiz_attempts = delivery.quiz_submissions.all().order_by(
                        "submitted_at"
                    )
                    attempt = None
                    if schedule_no == 1:
                        attempt = quiz_attempts.first()
                        attempt_number = 1
                    elif schedule_no > 1:
                        attempt_number = schedule_no
                        attempt = quiz_attempts[attempt_number - 1 :].first()
                    if attempt:
                        events.append(
                            Event(
                                type=EventType.QUIZ_SUBMITED,
                                timestamp=attempt.submitted_at,
                                event_data=QuizSubmitedEvent(
                                    quiz_id=delivery.course_content.quiz.id,  # type: ignore[union-attr]
                                    quiz_title=delivery.course_content.quiz.title,  # type: ignore[union-attr]
                                    score=attempt.score,
                                    is_passed=attempt.is_passed,
                                    attempt_number=attempt_number,
                                ),
                            )
                        )
        if (
            enrollment.status == EnrollmentStatus.COMPLETED
            and enrollment.final_state_at
        ):
            events.append(
                Event(
                    type=EventType.COURSE_COMPLETED,
                    timestamp=enrollment.final_state_at,
                    event_data=None,
                )
            )
        elif (
            enrollment.status == EnrollmentStatus.DEACTIVATED
            and enrollment.final_state_at
        ):
            events.append(
                Event(
                    type=EventType.DEACTIVATED,
                    timestamp=enrollment.final_state_at,
                    event_data=DeactivatedEvent(reason=enrollment.deactivation_reason),  # type: ignore[arg-type]
                )
            )

        return EnrollmentResponse.model_validate(
            {
                "id": enrollment.id,
                "learner": enrollment.learner,
                "course": enrollment.course,
                "status": enrollment.status,
                "events": events,
            }
        )


class LearnerDetailResponse(BaseModel):
    id: int
    email: str
    enrollments: list[EnrollmentSummaryResponse]

    model_config = ConfigDict(from_attributes=True)


class CreateCourseContentRequest(BaseModel):
    priority: int | None = Field(gt=0, examples=[1], default=None)
    waiting_period: WaitingPeriod
    content: LessonCreate | QuizCreate = Field(discriminator="type")

    @property
    def required_priority(self) -> int:
        if self.priority is not None:
            return self.priority
        else:
            raise ValueError("Priority must be set before converting to Django model.")

    def to_django_model(self, course: Course) -> CourseContent:
        lesson = None
        quiz = None
        if isinstance(self.content, LessonCreate):
            lesson = Lesson(
                title=self.content.title,
                content=self.content.content,
            )
            lesson.save()
            content_type = "lesson"
        elif isinstance(self.content, QuizCreate):
            quiz = Quiz(
                title=self.content.title,
                required_score=self.content.required_score,
                selection_strategy=self.content.selection_strategy.value,  # type: ignore[misc]
                deadline_days=self.content.deadline_days,  # type: ignore[misc]
            )
            quiz.save()
            for question_data in self.content.questions:
                question = Question(
                    text=question_data.text,
                    priority=question_data.priority,
                    quiz=quiz,
                )
                question.save()
                for answer_data in question_data.answers:
                    answer = Answer(
                        text=answer_data.text,
                        is_correct=answer_data.is_correct,
                        question=question,
                    )
                    answer.save()
            content_type = "quiz"
        course_content = CourseContent.objects.create(
            course=course,
            priority=self.required_priority,
            waiting_period=self.waiting_period.to_seconds(),
            lesson=lesson,
            quiz=quiz,
            type=content_type,
        )

        return course_content


class UpdateCourseContentRequest(BaseModel):
    priority: Optional[int] = Field(gt=0, examples=[1], default=None)
    waiting_period: Optional[WaitingPeriod] = None
    lesson: Optional[LessonUpdate] = None
    quiz: Optional[UpdateQuiz] = None
    is_published: Optional[bool] = None

    model_config = ConfigDict(extra="forbid")

    @model_validator(mode="after")
    def check_at_least_one(self) -> "UpdateCourseContentRequest":
        # Check if all fields are None
        fields = [
            self.priority,
            self.waiting_period,
            self.lesson,
            self.quiz,
            self.is_published,
        ]
        if not any(f is not None for f in fields):
            raise ValueError(
                "At least one of 'priority', 'waiting_period', 'lesson', 'quiz', or 'is_published' must be provided."
            )
        return self


class CourseContentResponse(BaseModel):
    id: int
    priority: int
    waiting_period: int
    type: str
    lesson: Optional[LessonResponse] = None
    quiz: Optional[QuizResponse] = None
    is_published: bool

    @field_serializer("waiting_period")
    def serialize_waiting_period(self, waiting_period: int) -> dict:
        return WaitingPeriod.from_seconds(waiting_period).model_dump()

    model_config = ConfigDict(from_attributes=True)


class CourseContentSummaryResponse(BaseModel):
    id: int
    title: str
    priority: int
    waiting_period: int
    is_published: bool
    type: str

    @field_serializer("waiting_period")
    def serialize_waiting_period(self, waiting_period: int) -> dict:
        return WaitingPeriod.from_seconds(waiting_period).model_dump()

    model_config = ConfigDict(from_attributes=True)


class ReorderCourseContentsRequest(BaseModel):
    ordered_content_ids: list[int] = Field(min_length=2, examples=[[3, 1, 2]])



from django.http import JsonResponse
from django.views import View
from django.urls import reverse
from django_email_learning.personalised.api.serializers import (
    QuizSubmissionRequest,
    QuestionResponse,
)
from django_email_learning.services.metrics_service import MetricsService
from django_email_learning.services import jwt_service
from django.utils.translation import gettext as _
from django_email_learning.models import (
    ContentDelivery,
    Enrollment,
    Certificate,
    QuizSubmission,
    Quiz,
    EnrollmentStatus,
)
from pydantic import ValidationError
import json
import logging

METRIC_SERVICE = MetricsService()

logger = logging.getLogger(__name__)


class QuizSubmissionView(View):
    def post(self, request, *args, **kwargs):  # type: ignore[no-untyped-def]
        payload = json.loads(request.body)
        try:
            serializer = QuizSubmissionRequest.model_validate(payload)
        except ValidationError as ve:
            return JsonResponse({"error": ve.errors()}, status=400)

        token = serializer.token
        answers = serializer.answers

        try:
            decoded = jwt_service.decode_jwt(token=token)
        except jwt_service.InvalidTokenException as jde:
            return JsonResponse({"error": str(jde)}, status=400)
        except jwt_service.ExpiredTokenException as ete:
            return JsonResponse({"error": str(ete)}, status=410)

        delivery_id = decoded["delivery_id"]

        try:
            delivery = ContentDelivery.objects.get(
                id=delivery_id, hash_value=decoded["delivery_hash"]
            )
        except ContentDelivery.DoesNotExist:
            return JsonResponse(
                {
                    "error": "The content delivery associated with this token does not exist."
                },
                status=500,
            )

        enrollment = delivery.enrollment
        if enrollment.status != EnrollmentStatus.ACTIVE:
            return JsonResponse({"error": "Quiz is not valid anymore"}, status=400)

        quiz = delivery.course_content.quiz
        if not quiz:
            return JsonResponse(
                {"error": "No quiz associated with this link"}, status=500
            )

        try:
            score, passed = self.calculate_score_and_passed(
                quiz, answers, decoded.get("question_ids")
            )
            logger.info(
                f"Learner ID {enrollment.learner.id} submitted quiz for Course {enrollment.course.title} with score {score}. Passed: {passed}"
            )
        except ValueError as ve:
            return JsonResponse({"error": str(ve)}, status=500)

        QuizSubmission.objects.create(
            delivery=delivery,
            score=score,
            is_passed=passed,
        )
        # Updating the ContentDelivery hash value to invalidate the quiz link
        delivery.update_hash()

        if passed:
            message = _("Congratulations! You have passed the quiz.")
            delivery = delivery.schedule_next_delivery()
            if not delivery:
                enrollment.graduate()
        else:
            # Check if it's the second attempt failing
            failed_submissions_count = QuizSubmission.objects.filter(
                delivery=delivery,
                is_passed=False,
            ).count()

            if failed_submissions_count > 1:
                message = _(
                    "You have failed the quiz twice. Unfortunately, you cannot continue this course with this enrollment. You can enroll again to retake the course."
                )
                logger.info(
                    f"Learner ID {enrollment.learner.id} has failed the quiz twice for Course {enrollment.course.title}. "
                    f"Marking enrollment as failed."
                )
                enrollment.fail()
            else:
                message = _(
                    "You have failed the quiz. You will receive another chance to retake it tomorrow."
                )
                logger.info(
                    f"Learner ID {enrollment.learner.id} has failed the quiz for Course {enrollment.course.title}. "
                    f"Scheduling a retry for the next day."
                )
                delivery.repeat_delivery_in_days(1)

        METRIC_SERVICE.quiz_submitted(
            course_slug=enrollment.course.slug,
            organization_id=enrollment.course.organization.id,
            quiz_id=quiz.id,
            is_passed=passed,
        )
        return JsonResponse(
            {
                "score": score,
                "passed": passed,
                "required_score": quiz.required_score,
                "message": message,
            },
            status=200,
        )

    @staticmethod
    def calculate_score_and_passed(
        quiz: Quiz, answers: list[QuestionResponse], question_ids: list | None
    ) -> tuple[int, bool]:
        # Optimize: Prefetch related answers to avoid N+1 queries
        questions = quiz.questions.prefetch_related("answers").all()

        if question_ids is None:
            question_ids = list(questions.values_list("id", flat=True))

        # Create lookup dictionaries for O(1) access
        questions_dict = {q.id: q for q in questions}
        answers_dict = {}
        correct_answers_count = {}

        # Pre-populate answer lookup and count correct answers per question
        for question_obj in questions:
            answers_dict[question_obj.id] = {
                a.id: a for a in question_obj.answers.all()
            }
            correct_answers_count[question_obj.id] = question_obj.answers.filter(
                is_correct=True
            ).count()

        base_score = 0.0

        for response in answers:
            if response.id not in question_ids:
                raise ValueError(
                    f"Question ID {response.id} is not valid for this quiz."
                )

            question = questions_dict.get(response.id)
            if not question:
                raise ValueError(f"Question ID {response.id} not found.")

            for answer_id in response.answers:
                # Check if answer exists for this question
                if answer_id not in answers_dict[response.id]:
                    raise ValueError(
                        f"Answer ID {answer_id} is not valid for Question ID {response.id}."
                    )

                answer = answers_dict[response.id][answer_id]
                correct_count = correct_answers_count[response.id]

                if answer.is_correct:
                    base_score += 1 / correct_count  # Full point for correct answer
                else:
                    base_score -= 0.5 / correct_count  # Penalty for incorrect answer

        score = round(base_score / len(question_ids) * 100)  # Score as percentage
        score = max(0, score)
        passed = score >= quiz.required_score
        return score, passed


class SubmitCertificateFormView(View):
    def post(self, request, *args, **kwargs):  # type: ignore[no-untyped-def]
        payload = json.loads(request.body)
        token = payload.get("token")
        name = payload.get("name")

        if not name:
            return JsonResponse({"error": _("Name is required")}, status=400)

        try:
            decoded = jwt_service.decode_jwt(token=token)
        except jwt_service.InvalidTokenException as jde:
            return JsonResponse({"error": str(jde)}, status=400)
        except jwt_service.ExpiredTokenException as ete:
            return JsonResponse({"error": str(ete)}, status=410)

        enrollment_id = decoded["enrollment_id"]

        try:
            enrollment = Enrollment.objects.get(id=enrollment_id)
        except Enrollment.DoesNotExist:
            return JsonResponse(
                {"error": "The enrollment associated with this token does not exist."},
                status=500,
            )

        if enrollment.status != EnrollmentStatus.COMPLETED:
            return JsonResponse(
                {
                    "error": "The enrollment is not completed. Certificate cannot be issued."
                },
                status=400,
            )
        certificate, created = Certificate.objects.get_or_create(
            enrollment=enrollment, defaults={"name_on_certificate": name}
        )
        certificate_path = reverse(
            "django_email_learning:personalised:certificate",
            kwargs={"certificate_number": certificate.certificate_number},
        )
        absolute_certificate_url = request.build_absolute_uri(certificate_path)

        return JsonResponse({"certificate_url": absolute_certificate_url}, status=200)



from django.conf import settings
from django.utils import timezone
import datetime
import jwt

SECRET = settings.SECRET_KEY
ALGORITHM = "HS256"


class InvalidTokenException(Exception):
    pass


class ExpiredTokenException(Exception):
    pass


def generate_jwt(
    payload: dict,
    expiration_seconds: int = 3600,
    exp: datetime.datetime | None = None,
) -> str:
    payload_copy = payload.copy()
    if not exp:
        exp = timezone.now() + datetime.timedelta(seconds=expiration_seconds)
    payload_copy["exp"] = exp
    token = jwt.encode(payload_copy, SECRET, algorithm=ALGORITHM)
    return token


def decode_jwt(token: str) -> dict:
    try:
        decoded = jwt.decode(token, SECRET, algorithms=[ALGORITHM])
        return decoded
    except (jwt.InvalidSignatureError, jwt.DecodeError, jwt.InvalidAlgorithmError):
        raise InvalidTokenException("The signature is invalid")
    except jwt.ExpiredSignatureError:
        raise ExpiredTokenException("The token is not valid anymore")



from django_email_learning.services.command_models.command_request import CommandRequest
from django_email_learning.services.command_models.enroll_command import EnrollCommand
from django_email_learning.services.command_models.unsubscribe_command import (
    UnsubscribeCommand,
)
from pydantic import ValidationError


class InvalidCommandError(Exception):
    pass


class CommandHandlerService:
    def handle_command(self, command: EnrollCommand | UnsubscribeCommand) -> None:
        try:
            request = CommandRequest(command=command)
            request.command.execute()
        except ValidationError as e:
            raise InvalidCommandError("Invalid command") from e

    def handle_json_command(self, json_command: dict) -> None:
        try:
            request = CommandRequest.model_validate(json_command)
            request.command.execute()
        except ValidationError as e:
            raise InvalidCommandError("Invalid command JSON") from e



# Generated by Django 6.0.3 on 2026-03-09 10:27

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Session",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("session_id", models.CharField(max_length=255, unique=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("jwt_token", models.TextField()),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("PENDING", "Pending"),
                            ("PROCESSING", "Processing"),
                            ("COMPLETED", "Completed"),
                            ("FAILED", "Failed"),
                        ],
                        default="PENDING",
                        max_length=255,
                    ),
                ),
            ],
        ),
    ]
